"""This module contains functions providing old-school static HTML output for the integrated visualization of binding
pockets obtained by the pocket miner.

:Deprecated: This module has been deprecated by the atligator web interface. See project atligator_web

:Author: Josef Kynast <josef.kynast@uni-bayreuth.de>
:Author: Felix Schwaegerl <felix.schwaegerl@uni-bayreuth.de>
:date: 2018-06-06
"""
import datetime
from pathlib import Path
from typing import List, Dict

from Bio.SeqUtils import seq1

from atligator.atlas import Atlas
from atligator.pocket_miner import Pocket


def generate_html_pocket_index(output_folder: str, atlas: Atlas, pockets: Dict[str, List[Pocket]]) -> None:
    """Generates a HTML index page showing summarizing information on the atlas and the mined pockets. The page also
    contains links to individual pocket descriptions which are supposed to be generated by the function
    generate_html_pocket_description.
    :param output_folder: where the index.html file is generated
    :param atlas: the atlas the pockets were mined from
    :param pockets: the mined pockets
    """
    html: List[str] = list()
    html.append('<!DOCTYPE html>')
    html.append('<html lang="en">')
    html.append('<head>')
    html.append('    <meta charset="utf-8">')
    html.append('</head>')
    html.append('</body>')
    html.append('    <h1>Atlas Binding Pockets</h1>')
    html.append(f"    <p>generated by ATLIGATOR's pocket miner script on {datetime.datetime.now()}.</p>")
    html.append('    <h2>General Information</h2>')
    html.append(f"    <p>number of datapoints in the atlas: {len(atlas.datapoints)}.</p>")
    html.append('    <h2>Binding Pockets</h2>')
    html.append('    <p>select one of the ligand residue types below:</p>')
    html.append('    <ul>')
    for lig_restype, pocketlist in sorted(pockets.items(), key=lambda e: e[0]):
        html.append(f'        <li><a href={lig_restype}.html>{lig_restype}</a> ({len(pocketlist)} pockets '
                    f'from {len([d for d in atlas.datapoints if d.ligand_restype == lig_restype])} datapoints)</li>')
    html.append('    </ul>')
    html.append('</body>')
    html.append('</html>')
    with open(Path(output_folder) / 'index.html', 'w') as fo:
        for line in html:
            fo.write(line + '\n')


def generate_html_pocket_description(output_folder: str, ligand_restype: str, pockets: List[Pocket],
                                     visualize_raw_pdb: bool, visualize_clustered_pdb: bool) -> None:
    """Generates a detailed HTML page showing pockets connected to a specfic ligand residue type. Optionally, the
    superimposed and clustered PDB structures are presented in a NGL.js viewer. The name of the generated file depends
    on the ligand residue type, e.g. LYS.html
    :param output_folder: will contain the generated file.
    :param ligand_restype: three-letter code of the residue type
    :param pockets: list of pockets belonging to the residue type
    :param visualize_raw_pdb: whether to visualize the superimposed PDB structure
    :param visualize_clustered_pdb: whether to visualize the clustered PDB structure
    """
    html: List[str] = list()
    html.append('<!DOCTYPE html>')
    html.append('<html lang="en">')
    html.append('<head>')
    html.append('    <meta charset="utf-8">')
    html.append('</head>')
    html.append('</body>')
    html.append(f'    <h3>{ligand_restype} Pockets</h3>')
    if visualize_raw_pdb or visualize_clustered_pdb:
        html.append('    <script src="https://unpkg.com/ngl"></script>')
        html.append(f'    <p>in the visualizations below, C atoms of the ligand are presented in white,'
                    f' of the binder in gray, respectively. CA atoms are in green.</p>')
        html.append(f'    <p>hint: double-click onto the structures for full-screen representation.</p>')
    for i, pocket in enumerate(pockets):
        pocket_code = f"{seq1(ligand_restype)}_{pocket.pocket_id()}_{int(pocket.support * 100.0)}"
        html.append(f'    <h4>Pocket {i + 1}: {pocket.pocket_id()}</h4>')
        html.append(f'    <p>from {pocket.n_ligand_origins()} datapoints; support: {pocket.support:.3f}</p>')
        if visualize_raw_pdb:
            html.append('    <p>superimposed pocket representation:</p>')
            html.append('    <script>')
            html.append('        document.addEventListener("DOMContentLoaded", function () {')
            html.append(f'            var stage = new NGL.Stage("viewport-{pocket_code}");')
            html.append('            stage.viewer.container.addEventListener("dblclick", function () {')
            html.append('                stage.toggleFullscreen();')
            html.append('            });')
            html.append('            var colorScheme = NGL.ColormakerRegistry.addSelectionScheme([')
            html.append('                ["darkgreen", ".CA"],')
            html.append('                ["white", ":A and _C"],')
            html.append('                ["gray", "(not :A) and _C"],')
            html.append('                ["red", "_O"],')
            html.append('                ["blue", "_N"],')
            html.append('                ["yellow", "_S"],')
            html.append('                ["green", "*"]')
            html.append(f'            ], "{pocket_code}");')
            html.append(f'            stage.loadFile("{pocket_code}.pdb").then(function (o) ' + '{')
            html.append('                o.addRepresentation("ball+stick", {color: colorScheme});')
            html.append('                o.autoView();')
            html.append('            });')
            html.append('        });')
            html.append('    </script>')
            html.append(f'    <div id="viewport-{pocket_code}" style="width:400px; height:300px;"></div>')
        if visualize_clustered_pdb:
            html.append('    <p>pocket centroid representation:</p>')
            html.append('    <script>')
            html.append('        document.addEventListener("DOMContentLoaded", function () {')
            html.append(f'            var stage = new NGL.Stage("viewport-{pocket_code}_clustered");')
            html.append('            stage.viewer.container.addEventListener("dblclick", function () {')
            html.append('                stage.toggleFullscreen();')
            html.append('            });')
            html.append('            var colorScheme = NGL.ColormakerRegistry.addSelectionScheme([')
            html.append('                ["darkgreen", ".CA"],')
            html.append('                ["white", ":A and _C"],')
            html.append('                ["gray", "(not :A) and _C"],')
            html.append('                ["red", "_O"],')
            html.append('                ["blue", "_N"],')
            html.append('                ["yellow", "_S"],')
            html.append('                ["green", "*"]')
            html.append(f'            ], "{pocket_code}_clustered");')
            html.append(f'            stage.loadFile("{pocket_code}_clustered.pdb").then(function (o) ' + '{')
            html.append('                o.addRepresentation("ball+stick", {color: colorScheme});')
            html.append('                o.autoView();')
            html.append('            });')
            html.append('        });')
            html.append('    </script>')
            html.append(f'    <div id="viewport-{pocket_code}_clustered" style="width:400px; height:300px;"></div>')
    html.append('    </ul>')
    html.append('    <p><a href="index.html">back to index</a></p>')
    html.append('</body>')
    html.append('</html>')
    with open(Path(output_folder) / f"{ligand_restype}.html", 'w') as fo:
        for line in html:
            fo.write(line + '\n')



"""Data structures and operations for handling geometric representations relative to different internal or external
coordinate systems.

:Author: Josef Kynast <josef.kynast@uni-bayreuth.de>
:Author: Felix Schwaegerl <felix.schwaegerl@uni-bayreuth.de>
:date: 2018-03-05
"""

from typing import List

from Bio.PDB.vectors import Vector
from numpy.linalg import inv

# typedefs
Matrix = List[List[float]]


class InternalCoordinates:
    """Instances of this class represent an internal coordinate system, which is defined by means of a translation
    vector and a rotation matrix."""

    def __init__(self, translation_vector: Vector, rotation_matrix: Matrix):
        """creates an instance of an internal coordinate system. It is defined by a translation vector and a rotation
        matrix.
        :param translation_vector vector between external and internal origin
        :param rotation_matrix matrix describing rotation from external to internal coordinates. This matrix should
        be normalized.
        """
        self.translation_vector = translation_vector
        self.rotation_matrix = rotation_matrix

    def external_to_internal(self, external_vector: Vector, direction: bool) -> Vector:
        """transforms a given external vector into a corresponding vector relative to the internal coordinate system.
        :param external_vector: vector in external coordinates to convert
        :param direction: if false, both translation and rotation will be applied; otherwise, only rotation
        :return: the vector in internal coordinates
        """
        if direction:
            internal_vector = external_vector
        else:
            internal_vector = external_vector - self.translation_vector
        internal_vector = internal_vector.left_multiply(self.rotation_matrix)
        return internal_vector

    def internal_to_external(self, internal_vector: Vector, direction: bool) -> Vector:
        """transforms a given vector relative to this internal coordinate system into a corresponding external
        (absolute) vector.
        :param internal_vector: vector to convert in internal coordinates
        :param direction: if true, both rotation and translation will be applied; otherwise, only rotation
        :return: the vector in external coordinates
        """
        external_vector = internal_vector.left_multiply(inv(self.rotation_matrix))
        if not direction:
            external_vector = external_vector + self.translation_vector
        return external_vector



"""This module provides data strucures and algoriths for mining frequent host pockets binding to specific ligand
residue types based on atlas data. It is based on association rule mining and frequent itemset mining algorithms.

:Author: Felix Schwaegerl <felix.schwaegerl@uni-bayreuth.de>
:date: 2018-04-20
"""
import json
import os
from concurrent.futures import ThreadPoolExecutor
from copy import copy
from math import sqrt
from random import sample, choice
from typing import List, Dict, Tuple, Set, Callable, TextIO, Union, Any

from Bio.PDB import Vector
from Bio.PDB.Chain import Chain
from Bio.PDB.Model import Model
from Bio.PDB.Structure import Structure
from Bio.SeqUtils import seq1
from scipy.optimize import linear_sum_assignment

from atligator.atlas import Atlas, AtlasDatapoint, AtlasEncoder
from atligator.pdb_util import canonical_amino_acids, find_scop_i_from_pdbs


class PocketItemset:
    """A pocket itemset is a collection of binder residues found to be interacting with a specific ligand residue."""
    items: Dict[str, int]

    def __init__(self, items: Dict[str, int] or str):
        """Creates a new instance.
        :param items: relates the types of items (i.e. their binder residue type) with the number of occurrences in
        a pocket. 'XXX' is a shorthand for the dict {'XXX': 1}
        """
        if type(items) is str:
            self.items = {items: 1}
        else:
            self.items = items

    def pocket_id(self) -> str:
        """
        :return: a sequence of one-letter codes of items part of the atlas. Multiple containment is expressed by
        duplication of letters. Example: 'KRRQNSS'
        """
        result = ''
        for k, v in self.items.items():
            for _ in range(v):
                result += seq1(k)
        return result

    def __contains__(self, other: 'PocketItemset') -> bool:
        """
        :param other:
        :return: whether this pocket itemset contains all items (and as least as many of duplicates) as other.
        """
        for itype, icount in other.items.items():
            if itype not in self.items or self.items[itype] < icount:
                return False
        return True

    def __add__(self, other: 'PocketItemset') -> 'PocketItemset':
        """
        :param other:
        :return: pocket itemset that represents the union of this and other.
        """
        u_items: Dict[str, int] = {}
        for itype in list(self.items.keys()) + list(other.items.keys()):
            icount_self = 0 if itype not in self.items else self.items[itype]
            icount_other = 0 if itype not in other.items else other.items[itype]
            u_items[itype] = icount_self + icount_other
        return PocketItemset(u_items)

    def __eq__(self, other) -> bool:
        return self.items == other.items

    def __hash__(self) -> int:
        return hash(frozenset(self.items.items()))

    def __len__(self) -> int:
        return sum(v for v in self.items.values())

    def __str__(self) -> str:
        return str(self.items)

    def __repr__(self) -> str:
        return str(self)


class PocketItemDataset:
    """A pocket item dataset organizes a collection of pocket itemsets and allows to make statistical statements
    (e.g., support, confidence). Furthermore, it enables the identification of frequent itemsets."""

    def __init__(self, itemsets: List[PocketItemset]):
        """Creates a new instance of the pocket dataset.
        :param itemsets: the initial collection of pocket itemsets
        """
        self.itemsets = itemsets
        self._support_cache: Dict[PocketItemset, float] = {}

    def implied_itemsets(self, itemset: PocketItemset) -> List[PocketItemset]:
        """Retrieves all itemsets managed by this dataset which contain the items described by the parameter itemset.
        :param itemset:
        """
        return list([i for i in self.itemsets if itemset in i])

    def support(self, itemset: PocketItemset) -> float:
        """The support of an itemset is defined as the ratio between the number of those itemsets that imply it and
        the number of itemsets managed in total.
        For performance reason, returned values are cached in the _support_cache, which is transparent to the caller.
        :param itemset:
        :return: the support of itemset. Ranges between 0 and 1.
        """
        if itemset in self._support_cache:
            return self._support_cache[itemset]
        else:
            lis = len(self.itemsets)
            support = 0.0 if lis == 0 else float(len(self.implied_itemsets(itemset))) / float(lis)
            self._support_cache[itemset] = support
            return support

    def confidence(self, premise: PocketItemset, conclusion: PocketItemset):
        """The confidence of an association rule X->Y between two itemsets is defined as the support of the union of X
        and Y, divided by the support of X.
        :param premise: left hand side of the association rule, i.e., X
        :param conclusion: right hand side, Y
        :return: the confidence of X->Y. Ranges between 0 and 1.
        """
        supp_pre = self.support(premise)
        return 0.0 if supp_pre == 0.0 else self.support(premise + conclusion) / supp_pre

    def find_associations(self, itemset: PocketItemset, confidence_threshold: float) -> Dict[str, float]:
        """Extends the given itemset by every possible item, taken from the list of 20 canonical amino acids. Relates
        the added item to the confidence related to the addition of the new item
        :param itemset: the base pocket itemset to find new association rules for
        :param confidence_threshold: associations showing a confidence below this value will not be returned
        :return: dict that maps potential residue types to the confidence predicted
        """
        associations: Dict[str, float] = {}
        # TODO alternative aas -> atm: associations related to binder residues, thus, not needed
        for restype in canonical_amino_acids:
            confidence = self.confidence(itemset, PocketItemset(restype))
            if confidence > confidence_threshold:
                associations[restype] = confidence
        return associations

    def find_frequent_itemsets(self, confidence_threshold: float, support_threshold: float,
                               min_itemset_cardinality: int,
                               cardinality_base: float = 2.0) -> List[Tuple[float, PocketItemset]]:
        """Uses a breadth-first search to internally create a frequent itemset lattice based on repeated applications
        of the find_associations function. Returns frequent itemsets associated with their support.
        :param confidence_threshold: association rules that fall below this threshold are not applied
        :param support_threshold: itemsets below this value are not considered as frequent
        :param min_itemset_cardinality: itemsets whose cardinality is equal or greater than this value are included in
        the result
        :param cardinality_base: used for fitness calculation of each itemset in 'support * sqrt(base)^|itemset|'.
        Higher values result in pockets with more items.
        :return: a list of tuples relating support values to corresponding itemsets mined from this dataset
        """
        result: List[Tuple[float, PocketItemset]] = []
        tabu: Set[PocketItemset] = set()
        previous: List[PocketItemset] = [PocketItemset({})]
        while len(previous) > 0:
            current = []
            for iset in previous:
                for new_item in self.find_associations(iset, confidence_threshold).keys():
                    new_itemset = iset + PocketItemset(new_item)
                    if new_itemset not in tabu:
                        tabu.add(new_itemset)
                        new_supp = self.support(new_itemset)
                        if new_supp > support_threshold:
                            current.append(new_itemset)
                            if len(new_itemset) >= min_itemset_cardinality:
                                result.append((new_supp, new_itemset))
            previous = current
        return sorted(result, key=lambda e: e[0] * float(sqrt(cardinality_base) ** float(len(e[1]))), reverse=True)


def extract_pocket_item_datasets(atlas: Atlas) -> Dict[str, PocketItemDataset]:
    """Evaluates an atlas and returns up to 20 pocket item datasets, one for each residue type occurring in the atlas.
    :param atlas: the input for the analysis
    :return: dict from ligand residue type to their individual item dataset
    """
    item_datasets: Dict[str, PocketItemDataset] = {}
    for lig_origin, datapoint_group in atlas.group_by_ligand_origin().items():
        itemset = PocketItemset({})
        for datapoint in datapoint_group:
            itemset += PocketItemset(datapoint.binder_restype)
        lig_restype = datapoint_group[0].ligand_restype
        if lig_restype not in item_datasets:
            item_datasets[lig_restype] = PocketItemDataset([])
        item_datasets[lig_restype].itemsets.append(itemset)
    return item_datasets


def filter_atlas_by_itemset(atlas: Atlas, ligand_restype: str, itemset: PocketItemset,
                            include_unrelated_residues: bool = False) -> Atlas:
    """Restricts a given atlas to those ligand residues whose binding pocket is represented by a given itemset. In
    this way, an atlas for a specific binding pocket can be obtained automatically.
    :param atlas: the atlas to filter
    :param ligand_restype: ligand residue type to assume for filtering. None if this should not be used as a criterion
    (in this case, the resulting atlas would still represent several ligand residue types)
    :param itemset: datapoint groups that match this itemset pass the filter
    :param include_unrelated_residues: if true, the resulting atlas also includes residue types that are not part of
    the actual pocket. This introduces more noise, but can also indicate interactions overlooked by itemset mining
    :return: the filtered atlas
    """
    filtered_dps: List[AtlasDatapoint] = []
    for _, datapoints in atlas.group_by_ligand_origin().items():
        g_lig_restype = datapoints[0].ligand_restype
        if ligand_restype is not None and g_lig_restype != ligand_restype:
            continue
        g_itemset = PocketItemset({})
        for datapoint in datapoints:
            g_itemset += PocketItemset(datapoint.binder_restype)
        if itemset in g_itemset:
            for datapoint in datapoints:
                if include_unrelated_residues or datapoint.binder_restype in itemset.items.keys():
                    filtered_dps.append(datapoint)
    return Atlas(filtered_dps)


def distance(res1: AtlasDatapoint, res2: AtlasDatapoint,
             distance_factor: float, orient_factor: float, secor_factor: float) -> float:
    """Distance function used for clustering of residues. Weighted sum of the distances between C_alpha vectors, and
     primary and secondary orientation angles.
    :param res1: first operand
    :param res2: second operand
    :param distance_factor: weight for the distance between c_alpha positions
    :param orient_factor: weight for the difference between primary orientation
    :param secor_factor: weight for the difference between secondary orientation
    :return: the distance calculated (a nonnegative scalar)
    """
    calpha_distance: float = abs((res1.binder_calpha() - res2.binder_calpha()).norm())
    orient_angle: float = abs(res1.binder_orientation().angle(res2.binder_orientation()))
    sec_orient_angle: float = abs(res1.binder_secondary_orientation().angle(res2.binder_secondary_orientation()))
    return distance_factor * calpha_distance + orient_factor * orient_angle + secor_factor * sec_orient_angle


class PocketCluster:
    """A cluster of atlas residues that represents a part of a pocket."""

    @classmethod
    def from_json(cls, cluster: Dict[str, Any]):
        """
        Return a PocketCluster from a json representation of one.
        :param cluster: json representation of a PocketCluster or open file handle containing it
        :return: PocketCluster based on json
        """
        centroid = AtlasDatapoint.from_json(cluster['centroid'])
        new_cluster = PocketCluster(ligand_restype=cluster['ligand_restype'],
                                    binder_restype=cluster['binder_restype'],
                                    initial_centroid=centroid)
        new_cluster.variance = cluster['variance']
        new_cluster.has_changed = cluster['has_changed']
        new_cluster.members = [AtlasDatapoint.from_json(mem) for mem in cluster['members']]
        new_cluster.future_members = [AtlasDatapoint.from_json(fmem) for fmem in cluster['future_members']]
        return new_cluster

    def __init__(self, ligand_restype: str, binder_restype: str, initial_centroid: AtlasDatapoint):
        """ Creates a new cluster.
        :param ligand_restype: same as for all clustered atlas residues
        :param binder_restype: same as for all clustered atlas residues
        :param initial_centroid: the initial centroid of the cluster. Not counted as a member.
        """
        self.ligand_restype = ligand_restype
        self.binder_restype = binder_restype
        self.members: List[AtlasDatapoint] = []
        self.centroid = copy(initial_centroid)
        self.variance = 0.0
        self.future_members: List[AtlasDatapoint] = []
        self.has_changed = True

    def __len__(self):
        return len(self.members)

    def get_centroid(self) -> AtlasDatapoint:
        """Calculates a new centroid based on the members of this clusters. It is obtained by calculating the
        arithmetic means of the components of all c_alpha, c_beta, and c_o vectors.
        :return: an artificial atlas residue that represents the centroid of this cluster
        """
        cent = AtlasDatapoint(self.ligand_restype, self.binder_restype, None, None, {}, {})
        cl = 1.0 / float(len(self.future_members))
        for ares in self.future_members:
            for lig_atom, lig_pos in ares.ligand_atoms.items():
                if lig_atom not in cent.ligand_atoms:
                    cent.ligand_atoms[lig_atom] = Vector(0.0, 0.0, 0.0)
                cent.ligand_atoms[lig_atom] += ares.ligand_atoms[lig_atom] ** cl
            for bin_atom, bin_pos in ares.binder_atoms.items():
                if bin_atom not in cent.binder_atoms:
                    cent.binder_atoms[bin_atom] = Vector(0.0, 0.0, 0.0)
                cent.binder_atoms[bin_atom] += ares.binder_atoms[bin_atom] ** cl
        return cent

    def update(self, distance_factor: float, orient_factor: float, secor_factor: float, keep_members: bool = False) \
            -> None:
        """To be called after every clustering iteration. Updates the centroid, the cached variance, saves the list
        of members in order to detect whether the contents have changed, and clears the current member list.
        :param distance_factor: weight for the distance between c_alpha positions in distance function
        :param orient_factor: weight for the difference between primary orientation in distance function
        :param secor_factor: weight for the difference between secondary orientation in distance function
        """
        if len(self.future_members) > 0:
            self.centroid = self.get_centroid()
        self.variance = self.get_variance(None, distance_factor, orient_factor, secor_factor)
        self.has_changed = frozenset(self.members) != frozenset(self.future_members)
        if keep_members:
            self.members = self.members + self.future_members[:]
        else:
            self.members = self.future_members[:]
        self.future_members.clear()

    def get_variance(self, additional_member: AtlasDatapoint or None,
                     distance_factor: float, orient_factor: float, secor_factor: float) -> float:
        """Obtains the variance of this cluster, which is obtained by the sum of squares of distances between all
        members and the centroid, divided by the number of members.
        :param additional_member: optionally, a hypothetical member can be provided here. It will be treated as if it
        were actually part of the cluster. Otherwise, fill this parameter with None.
        :param distance_factor: weight for the distance between c_alpha positions in distance function
        :param orient_factor: weight for the difference between primary orientation in distance function
        :param secor_factor: weight for the difference between secondary orientation in distance function
        :return: the variance of this cluster, a nonnegative scalar. Square of the standard deviation
        """
        if len(self.future_members) == 0 and additional_member is None:
            return 0.0
        dist = 0.0
        for ares in self.future_members + ([] if additional_member is None else [additional_member]):
            dist += (distance(self.centroid, ares, distance_factor, orient_factor, secor_factor) ** 2.0)
        return dist / float(len(self.future_members) + (0 if additional_member is None else 1))

    def get_variance_increase(self, candidate: AtlasDatapoint,
                              distance_factor: float, orient_factor: float, secor_factor: float) -> float:
        """Calculates the increase in variance that would be obtained by adding a given candidate to the list of members
        of this cluster.
        :param candidate: virtual additional member of the cluster (will not be added)
        :param distance_factor: weight for the distance between c_alpha positions in distance function
        :param orient_factor: weight for the difference between primary orientation in distance function
        :param secor_factor: weight for the difference between secondary orientation in distance function
        :return: variance including the candidate minus variance without the candidate. May be negative
        """
        return self.get_variance(candidate, distance_factor, orient_factor, secor_factor) - self.get_variance(
            None, distance_factor, orient_factor, secor_factor)

    def get_least_representative_member(self, distance_factor: float, orient_factor: float, secor_factor: float) \
            -> AtlasDatapoint:
        """
        :return: the member of the cluster that has the largest distance to the centroid.
        :param distance_factor: weight for the distance between c_alpha positions in distance function
        :param orient_factor: weight for the difference between primary orientation in distance function
        :param secor_factor: weight for the difference between secondary orientation in distance function
        """
        max_dist = 0.0
        lrm = None
        for m in self.members:
            dist = distance(self.centroid, m, distance_factor, orient_factor, secor_factor)
            if dist > max_dist:
                max_dist = dist
                lrm = m
        return lrm

    def get_most_representative_member(self, distance_factor: float, orient_factor: float, secor_factor: float) \
            -> AtlasDatapoint:
        """
        :return: the member of the cluster that has the shortest distance to the centroid.
        :param distance_factor: weight for the distance between c_alpha positions in distance function
        :param orient_factor: weight for the difference between primary orientation in distance function
        :param secor_factor: weight for the difference between secondary orientation in distance function
        """
        min_dist = float('inf')
        mrm = None
        for m in self.members:
            dist = distance(self.centroid, m, distance_factor, orient_factor, secor_factor)
            if dist < min_dist:
                min_dist = dist
                mrm = m
        return mrm

    def get_dist_to_centroid(self, other: AtlasDatapoint, distance_factor: float,
                             orient_factor: float, secor_factor: float) -> float:
        return distance(self.centroid, other, distance_factor, orient_factor, secor_factor)


class Pocket:
    """A pocket is a collection of pocket clusters that have emerged from a given itemset."""

    @classmethod
    def from_json(cls, pocket):
        """
        Return a Pocket from a json representation of one.
        :param pocket: json representation of a Pocket
        :return: Pocket based on json
        """
        pocket_itemset = PocketItemset(items=pocket['itemset'])
        pocket_clusters = []
        for cluster in pocket['clusters']:
            pocket_clusters.append(PocketCluster.from_json(cluster))
        return Pocket(restype=pocket['restype'],
                      support=pocket['support'],
                      itemset=pocket_itemset,
                      clusters=pocket_clusters)

    def to_json(self):
        """
        Return a json representation of this Pocket.
        :return: json string based on Pocket
        """
        return json.dumps(self, cls=PocketsEncoder)

    def __init__(self, restype: str, itemset: PocketItemset, support: float, clusters: List[PocketCluster]):
        """
        :param restype: the ligand residue type represented by this pocket
        :param itemset: the itemset containing the binder residue types part of the binding pocket
        :param support: absolute support of the pocket's itemset in the dataset
        :param clusters: contains the positions of clustered atlas residues for every element of the itemset at
        corresponding positions.
        """
        self.restype = restype
        self.itemset = itemset
        self.support = support
        self.clusters = clusters

    def __repr__(self):
        return f"<Pocket {self.restype}: {self.itemset}>"

    def __len__(self) -> int:
        return self.n_ligand_origins()

    def pocket_id(self) -> str:
        """
        :return: an identifier for this pocket, based on its underlying dataset
        """
        return self.itemset.pocket_id()

    def n_ligand_origins(self) -> int:
        """
        :return: the number of distinct ligand origins the members of this cluster have emerged from
        """
        lig_ors: Set[str] = set()
        for c in self.clusters:
            for m in c.members:
                lig_ors.add(m.ligand_origin)
        return len(lig_ors)

    def origins(self, scop_db: str = None,
                stringify: bool = True, scop_only: bool = True) -> Dict or str or None:
        """
        Specfies origins of data points in this pocket (in scope classifiers)
        :param scop_db: Scope database reference
        :param stringify: If true the output will be a string
        :param scop_only: If true the output will be a dictionary with number scope occurances (or its string)
        :return: Either a dict of scope identifiers with number of occurances or
        a dict with pdb names and their correlated scope identifiers
        """

        def get_scop_i(pdb_scop_dict, sort_d: bool = True) -> Dict or List:
            """
            Returns number of occurances of this pocket in a specific SCOPe family
            :param pdb_scop_dict: Dictionary to find pdb - scop connections from
            :param sort_d: If true the output will be a list of keys and values of the output dictionary
            sorted by number of occurances
            :return: Dictionary with scope identifiers and the number of occurances of this pocket
            """
            scop_i_n: Dict[str, int] = {"None": 0}
            for l in pdb_scop_dict.values():
                if l is None:
                    scop_i_n["None"] += 1
                    continue
                for i in l:
                    if i not in scop_i_n.keys():
                        scop_i_n[i] = 1
                    else:
                        scop_i_n[i] += 1

            if not sort_d:
                return scop_i_n
            return [(k, scop_i_n[k]) for k in sorted(scop_i_n, key=scop_i_n.__getitem__, reverse=True)]

        if os.path.isfile(scop_db):
            pdb_set: Set[str] = set()
            for c in self.clusters:
                for m in c.members:
                    pdb_set.add(m.ligand_origin)
            pdbs: List[str] = []
            for pdb in pdb_set:
                pdbs.append(pdb[:4])
            d = find_scop_i_from_pdbs(pdbs, scop_db)
            if scop_only:
                return get_scop_i(d)
            return d
        return None

    def to_atlas(self) -> Atlas:
        """Converts a pocket into an atlas based on the members of the clusters.
        :return: an atlas that represents the pocket
        """
        a_residues: List[AtlasDatapoint] = []
        for c in self.clusters:
            a_residues.extend(c.members)
        return Atlas(a_residues)

    def get_most_representative_group(self, distance_factor: float, orient_factor: float, secor_factor: float):
        """ Selects the most representative group of cluster members. A group consists of a ligand residue and
        binder residues. The number of binder residues corresponds to the number of clusters and thus to the cardinality
        of this pocket.
        :param distance_factor: weight for the distance between c_alpha positions in distance function
        :param orient_factor: weight for the difference between primary orientation in distance function
        :param secor_factor: weight for the difference between secondary orientation in distance function
        :return: an atlas that represents the clustered pocket
        """
        min_len = 2 ** 10000
        min_c = None
        # TODO only applicable if all clusters must include at least one..
        for c_i, c in enumerate(self.clusters):  # Find smallest cluster first
            if len(c.members) < min_len:
                min_c = c_i
                min_len = len(c.members)
        suited_origins: Dict[str, List] = {}
        for member in self.clusters[min_c].members:  # Find the ligand origins that can be found in all clusters
            if all(member.ligand_origin in (m.ligand_origin for m in c.members) for c in self.clusters):
                suited_origins[member.ligand_origin] = [None for _ in range(len(self.clusters))]
        if not len(suited_origins):  # If no suited origin is found, no group can be assigned
            print(self.pocket_id())
            return None, None

        origin_sums = {}
        for so in suited_origins.keys():  # For each suited origins find the best representatives in each cluster
            for c_i, c in enumerate(self.clusters):
                minimum = float('inf')
                for m in c.members:
                    if m.ligand_origin == so:
                        dist = c.get_dist_to_centroid(m, distance_factor, orient_factor, secor_factor)
                        if dist < minimum:
                            suited_origins[so][c_i] = [m.binder_origin, dist]  # Save the binder origin
            origin_sums[so] = sum(d[1] for d in suited_origins[so])
        best_origin = min(origin_sums, key=origin_sums.get)
        return suited_origins[best_origin], best_origin

    def to_clustered_atlas(self, distance_factor: float, orient_factor: float, secor_factor: float,
                           most_represent_group: bool = False) -> Atlas:
        """Converts a pocket into an atlas based on the centroids of the clusters.
        :param distance_factor: weight for the distance between c_alpha positions in distance function
        :param orient_factor: weight for the difference between primary orientation in distance function
        :param secor_factor: weight for the difference between secondary orientation in distance function
        :return: an atlas that represents the clustered pocket
        """
        ca_residues: List[AtlasDatapoint] = []
        if most_represent_group:  # Select for best grouped binder residues within this itemset
            members_of_best_lig_origin, best_ligand_origin = \
                self.get_most_representative_group(distance_factor, orient_factor, secor_factor)
            if best_ligand_origin is None:
                return Atlas([])
            for c_i, c in enumerate(self.clusters):
                for member in c.members:
                    if member.binder_origin == members_of_best_lig_origin[c_i][0] \
                            and member.ligand_origin == best_ligand_origin:
                        ca_residues.append(member)
                        break

        else:
            for bin_restype, count in self.itemset.items.items():
                r_clusters = list([c for c in self.clusters if c.binder_restype == bin_restype])
                for c in r_clusters:
                    ca_residues.append(c.get_most_representative_member(distance_factor, orient_factor, secor_factor))
        return Atlas(ca_residues)

    def to_pdb_structure(self) -> Structure:
        """
        :return: A PDB compliant structure that contains the datapoints of this pocket. By convention, chain A contains
        different conformations of the ligand residues of all datapoints, chains B and further contain residue
        coordinates for the conformations represented by individual members of the clusters.
        """
        model = Model(0)
        resid = 1
        ligand_chain = Chain('A')
        ligand_origins: Set[str] = set()
        for cluster in self.clusters:
            for member in cluster.members:
                if member.ligand_origin in ligand_origins:
                    continue
                ligres = member.ligand_to_pdb_residue(resid)
                ligand_chain.add(ligres)
                ligand_origins.add(member.ligand_origin)
                resid += 1
        model.add(ligand_chain)
        for i, cluster in enumerate(self.clusters):
            binder_chain_id = chr(ord('B') + i)
            binder_chain = Chain(binder_chain_id)
            for member in cluster.members:
                binres = member.binder_to_pdb_residue(resid)
                binder_chain.add(binres)
                resid += 1
            model.add(binder_chain)
        struc = Structure(self.pocket_id())
        struc.add(model)
        return struc

    def to_clustered_pdb_structure(self, distance_factor: float, orient_factor: float, secor_factor: float,
                                   most_represent_group: bool = True) -> Structure:
        """
        :param distance_factor: weight for the distance between c_alpha positions in distance function
        :param orient_factor: weight for the difference between primary orientation in distance function
        :param secor_factor: weight for the difference between secondary orientation in distance function
        :param most_represent_group: If True, the best group of residues (originating from one pdb structure!) are
        selected. Otherwise, the best individual residues are selected even if the originate from different structures
        and might not be senseful in combination (sterical hindering etc.).

        :return: A PDB compliant structure that contains the data of this pocket in a clustered way. By convention,
        chain A contains the best-matching (i.e. closest to centroid) conformation of the ligand residues of all
        datapoints, chains B and further contain best-matching residue coordinates for the conformations represented
        by individual members of the clusters. Thus, every chain posesses one residue.
        """
        model = Model(0)
        resid = 1
        ligand_chain = Chain('A')
        repr_ligand_origins: Set[str] = set()
        if most_represent_group:  # Select for best grouped binder residues within this itemset
            members_of_best_lig_origin, best_ligand_origin = \
                self.get_most_representative_group(distance_factor, orient_factor, secor_factor)
            if best_ligand_origin is None:
                return Structure(self.pocket_id())
            for c in self.clusters:
                for member in c.members:
                    if member.ligand_origin == best_ligand_origin:
                        ligres = member.ligand_to_pdb_residue(resid)
                        ligand_chain.add(ligres)
                        resid += 1
                        break
                break
            chain_counter = 0
            for c_i, c in enumerate(self.clusters):
                for member in c.members:
                    if member.binder_origin == members_of_best_lig_origin[c_i][0] \
                            and member.ligand_origin == best_ligand_origin:
                        binres = member.binder_to_pdb_residue(resid)
                        binder_chain_id = chr(ord('B') + chain_counter)
                        binder_chain = Chain(binder_chain_id)
                        binder_chain.add(binres)
                        resid += 1
                        chain_counter += 1
                        model.add(binder_chain)
                        break
            model.add(ligand_chain)
            struc = Structure(self.pocket_id())
            struc.add(model)
        else:  # find the best fitting individual residues within this itemset
            for cluster in self.clusters:
                cluster_repr = cluster.get_most_representative_member(distance_factor, orient_factor, secor_factor)
                if cluster_repr is None or cluster_repr.ligand_origin in repr_ligand_origins:
                    continue
                ligres = cluster_repr.ligand_to_pdb_residue(resid)
                ligand_chain.add(ligres)
                repr_ligand_origins.add(cluster_repr.ligand_origin)
                resid += 1
                break
            for i, cluster in enumerate(self.clusters):
                cluster_repr = cluster.get_most_representative_member(distance_factor, orient_factor, secor_factor)
                if cluster_repr is None:
                    continue
                binder_chain_id = chr(ord('B') + i)
                binder_chain = Chain(binder_chain_id)
                binres = cluster_repr.binder_to_pdb_residue(resid)
                binder_chain.add(binres)
                resid += 1
                model.add(binder_chain)
            model.add(ligand_chain)
            struc = Structure(self.pocket_id())
            struc.add(model)
        return struc


def pockets_to_json_file(pockets: Dict[str, List[Pocket]], fp):
    """
    Dumps the json represenation of Pockets to an open file handle
    :param pockets: Dict with Pockets to convert to json
    :param fp: open file handle to write json into
    :return: None
    """
    return json.dump(pockets, fp=fp, cls=PocketsEncoder)


def pockets_to_json(pockets: Dict[str, List[Pocket]]):
    """
    Returns the json represenation of Pockets
    :param pockets: Dict with Pockets to convert to json
    :return: json represenation of Pockets
    """
    return json.dumps(pockets, cls=PocketsEncoder)


def json_to_pockets(json_str_or_file: Union[str, TextIO]):
    """
    Return a Pockets dict from a json representation of one.
    :param json_str_or_file: json representation of Pockets or open file handle containing it
    :return: dictionary with Pockets
    """

    def convert(pockets: Dict[str, List[Dict]]):
        return_pockets = {}
        for ligand_aa, ligand_pockets in pockets.items():
            return_pockets[ligand_aa] = []
            for pocket in ligand_pockets:
                return_pockets[ligand_aa].append(Pocket.from_json(pocket))
        return return_pockets

    if isinstance(json_str_or_file, str):
        return convert(json.loads(json_str_or_file))
    else:
        return convert(json.load(json_str_or_file))


def clusterize_pocket_atlas(atlas: Atlas, itemset: PocketItemset, support: float, variance_threshold: float,
                            distance_factor: float, orient_factor: float, secor_factor: float,
                            one_per_cluster: bool = True, more_than_one_ok: bool = True) -> Pocket:
    """Takes an atlas, which already represents an individual binding pocket, and an itemset this atlas must be based
    on. Calculates a list of pocket clusters, which non-disjointly divide up the residues of the input atlas. For every
    residue type, at least as many pocket clusters as reflected by the itemset are obtained.
    The implementation is based on Lloyd's k-means algorithm. In addition, a post-processing step ensures that the
    obtained clusters' variance is below a given variance threshold by removing outliers iteratively.
    :param atlas: the pocket atlas to clusterize
    :param itemset: the itemset the pocket atlas is based on
    :param support: support of the itemset
    :param variance_threshold: in case the calculated clusters display a variance above this threshold, they will be
    divided up hierarchically in an iterative post-processing step.
    :param distance_factor: weight for the distance between c_alpha positions in distance function
    :param orient_factor: weight for the difference between primary orientation in distance function
    :param secor_factor: weight for the difference between secondary orientation in distance function
    :return: a list of pocket clusters representing a non-disjoint decomposition of the pocket atlas
    """

    def __assign_to_best_cluster(candidate_clusters: List[PocketCluster], datapoint: AtlasDatapoint) -> None:
        """Utility function that finds the most suitable among a set of candidate clusters, such that the variance of
        the cluster is increased to the lowest possible extent. The given datapoint is then appended to this cluster.
        :param candidate_clusters: the set of clusters to find suitable candidates within
        :param datapoint: the datapoint to decide cluster membership for
        """
        min_vi = float('inf')
        min_cluster: PocketCluster = None
        for c in candidate_clusters:
            if c.binder_restype == datapoint.binder_restype:
                vi = c.get_dist_to_centroid(datapoint, distance_factor, orient_factor, secor_factor)
                if vi < min_vi:
                    min_vi = vi
                    min_cluster = c
        min_cluster.future_members.append(datapoint)

    def __get_restypes(datapoints: List[AtlasDatapoint]):
        restypes = {}
        for dp_ in datapoints:
            try:
                restypes[dp_.binder_restype] += 1
            except KeyError:
                restypes[dp_.binder_restype] = 1
        return restypes

    def __assign_one_by_cluster(candidate_clusters: List[PocketCluster], datapoints: List[AtlasDatapoint]) -> None:
        """
        Uses the Hungarian algorithm to assign one datapoint to one cluster each. This is crucial, if one pocket
        includes several residues of the same type and even more of the same type in one single set.
        Example:
        For Ligand Tyr there is a pocket called GAA which includes Gly and two Ala. In one structure a Tyr is
        interacting with three Alanines. Of these three two must be assigned to the two pocket clusters. The third one
        should be the most outlying one and is discarded.

        :param candidate_clusters: A list of the Clusters the datapoints should be assigned to
        :param datapoints: A list of the datapoints
        :return: None
        """
        restypes = __get_restypes(datapoints)
        for restype, number in restypes.items():
            res_dp_list = [x for x in datapoints if x.binder_restype == restype]
            dist_matrix: List[List[float]] = [[] for _ in range(len(res_dp_list))]
            for i, res_dp in enumerate(res_dp_list):
                for c in candidate_clusters:
                    if c.binder_restype == restype:
                        dist_matrix[i].append(
                            c.get_dist_to_centroid(res_dp, distance_factor, orient_factor, secor_factor))
            if number != itemset.items[restype]:
                maximum = max(max(dist_matrix)) + 1.0
                for i, res_dp in enumerate(res_dp_list):
                    for dummy_clusters in range(number - itemset.items[restype]):
                        dist_matrix[i].append(maximum)
            assignment = linear_sum_assignment(dist_matrix)
            count = 0
            for c in candidate_clusters:
                if c.binder_restype == restype:
                    for i, a in enumerate(assignment[1]):
                        if a == count:
                            c.future_members.append(res_dp_list[assignment[0][i]])
                            break
                    count += 1


    clusters: List[PocketCluster] = []
    # create one cluster per item. Use representatives of a random pocket of the atlas as initial centroids.
    random_dp = choice(atlas.datapoints)
    random_pocket = [dp for dp in atlas.datapoints if dp.ligand_origin == random_dp.ligand_origin]
    for binder_restype, count in itemset.items.items():
        dps = sample(list([dp for dp in random_pocket if dp.binder_restype == binder_restype]), count)
        for dp in dps:
            clusters.append(PocketCluster(dp.ligand_restype, binder_restype, dp))
    # apply k-means clustering iteratively as long as the assignment of elements to clusters changes
    while any(c.has_changed for c in clusters):
        if one_per_cluster:
            # assign only "number-of-clusters" datapoints from each ligand residue to closest cluster
            for dps in atlas.group_by_ligand_origin().values():
                __assign_one_by_cluster(clusters, dps)
        else:
            # assign each datapoint to that cluster with the lowest distance to centroid
            for dp in atlas.datapoints:
                __assign_to_best_cluster(clusters, dp)
        # update the centroids and has_changed of every cluster
        for c in clusters:
            c.update(distance_factor, orient_factor, secor_factor)
    if one_per_cluster and more_than_one_ok:
        for dp in atlas.datapoints:
            __assign_to_best_cluster(clusters, dp)
        for c in clusters:
            c.update(distance_factor, orient_factor, secor_factor)
    # postprocessing: Until the variance of any of the clusters no longer exceeds the threshold, remove the most
    # distant member from the cluster.
    for c in clusters:
        while c.variance > variance_threshold:
            lrm = c.get_least_representative_member(distance_factor, orient_factor, secor_factor)
            c.future_members = [m for m in c.members if m != lrm]
            c.update(distance_factor, orient_factor, secor_factor)
    return Pocket(atlas.datapoints[0].ligand_restype, itemset, support, clusters)


def mine_pockets(atlas: Atlas, max_pockets_per_ligand_restype: int = 5, min_itemset_cardinality: int = 2,
                 confidence_threshold: float = 0.05, support_threshold: float = 0.05, variance_threshold: float = 5.0,
                 distance_factor: float = 1.0, orient_factor: float = 2.0, secor_factor: float = 2.0,
                 n_workers: int = -1, itemset_observer: Callable[[int, int], None] = None,
                 clustering_observer: Callable[[int, int], None] = None, cardinality_base: float = 2.0
                 ) -> Dict[str, List[Pocket]]:
    """Facade function that extracts pocket item datasets from an atlas, performs a frequent itemset mining analysis
    specifically for every residue type, and applies clustering in order to identify pockets. These pockets are then
    returned, grouped by ligand residue type
    :param atlas: the collection of datapoints taken as basis for the analysis
    :param max_pockets_per_ligand_restype: the result will contain no more than this number of pockets per ligand
    residue type
    :param min_itemset_cardinality: frequent itemsets must have at least this cardinality. I.e., the resulting pockets
    will consists of at least this number of residues
    :param confidence_threshold: association rules that fall below this threshold are not applied
    :param support_threshold: itemsets below this value are not considered as frequent
    :param variance_threshold: if apply_clustering, the residues contained by every cluster are allowed to have at
    most this variance; if this is exceeded, the cluster is reduced in an iterative post-processing step
    :param distance_factor: weight for the distance between c_alpha positions in distance function
    :param orient_factor: weight for the difference between primary orientation in distance function
    :param secor_factor: weight for the difference between secondary orientation in distance function
    :param n_workers: number of parallel processes to exploit
    :param itemset_observer: watches the progress of itemset mining
    :param clustering_observer: watches the progress of clustering
    :param cardinality_base: used for fitness calculation of each itemset in 'support * sqrt(base)^|itemset|'.
    Higher values result in pockets with more items.
    :return: a dictionary that maps every ligand residue type to a collection of a pockets.
    confidence, and the filtered atlas.
    """
    if n_workers == -1:
        n_workers = os.cpu_count()

    def __filter_and_clusterize(_ligrestype: str, _itemset: PocketItemset, _support: float) -> Pocket:
        """
        worker function that filters and clusterizes the itemset based on the atlas specified in the parent function.
        :param _ligrestype:
        :param _itemset:
        :param _support:
        :return: the pocket calculated by the worker
        """
        _pocket_atlas = filter_atlas_by_itemset(atlas, _ligrestype, _itemset, False)
        _pocket = clusterize_pocket_atlas(_pocket_atlas, _itemset, _support, variance_threshold,
                                          distance_factor, orient_factor, secor_factor)
        return _pocket

    datasets = extract_pocket_item_datasets(atlas)
    lig_restypes: List[str] = []
    frequent_itemsets: List[PocketItemset] = []
    supports: List[float] = []
    datasets_items = datasets.items()
    total = len(datasets_items)
    processed = 0
    for lig_restype, dataset in datasets_items:
        dataset_itemsets: List[PocketItemset] = []
        items_added = 0
        for (support, itemset) in dataset.find_frequent_itemsets(confidence_threshold, support_threshold,
                                                                 min_itemset_cardinality, cardinality_base):
            if not any(itemset in i for i in dataset_itemsets):  # avoid that subsets of existing itemsets are returned
                lig_restypes.append(lig_restype)
                dataset_itemsets.append(itemset)
                supports.append(support)
                items_added += 1
                if items_added >= max_pockets_per_ligand_restype:
                    break
        processed += 1
        if itemset_observer is not None:
            itemset_observer(processed, total)
        frequent_itemsets.extend(dataset_itemsets)
    pockets: List[Pocket] = []
    total = len(lig_restypes)
    processed = 0
    with ThreadPoolExecutor(n_workers) as executor:
        for pocket in executor.map(__filter_and_clusterize, lig_restypes, frequent_itemsets, supports):
            pockets.append(pocket)
            processed += 1
            if clustering_observer is not None:
                clustering_observer(processed, total)
    result: Dict[str, List[Pocket]] = {}
    for pocket in pockets:
        if pocket.restype not in result:
            result[pocket.restype] = []
        result[pocket.restype].append(pocket)
    return result


class PocketsEncoder(AtlasEncoder):
    """
    Enables the json Encoder to also encode an PocketCluster, Pocket and PocketItemset
    """

    def default(self, obj):
        if isinstance(obj, PocketCluster):
            return {'ligand_restype': obj.ligand_restype,
                    'binder_restype': obj.binder_restype,
                    'members': obj.members,
                    'centroid': obj.centroid,
                    'variance': obj.variance,
                    'future_members': obj.future_members,
                    'has_changed': obj.has_changed}
        if isinstance(obj, PocketItemset):
            return {'items': obj.items}
        if isinstance(obj, Pocket):
            return {'restype': obj.restype,
                    'itemset': obj.itemset,
                    'support': obj.support,
                    'clusters': obj.clusters}
        else:
            return AtlasEncoder.default(self, obj)



"""This module contains data structures and operations for the handling of designed ligands and designed residues, based
on which the prediction of preferred residue types shall be undertaken.

:Author: Josef Kynast <josef.kynast@uni-bayreuth.de>
:Author: Felix Schwaegerl <felix.schwaegerl@uni-bayreuth.de>
:date: 2018-02-28
"""

from typing import List, Tuple, Dict

from Bio.PDB.Atom import Atom
from Bio.PDB.Structure import Structure
from Bio.PDB.vectors import Vector
from Bio.SeqUtils import seq1

from atligator.geometry import InternalCoordinates, Matrix
from atligator.pdb_util import is_amino_acid_residue, get_cbeta_position

# typedefs
LigandInfo = Tuple[str, str, str]  # (chain_id, residue_id, new_residue_type)
BinderInfo = Tuple[str, str]  # (chain_id, residue_id)
CouplingInfo = Tuple[str, List[str]]  # (chain_id, {residue_id})


class AbstractResidue:
    """A residue is defined by a dictionary of atoms in absolute coordinates."""

    def __init__(self, atoms: Dict[str, Vector]):
        """Creates a new residue.
        :param atoms
        """
        self.atoms = atoms

    def calpha(self) -> Vector:
        return self.atoms['CA']

    def cbeta(self) -> Vector:
        return self.atoms['CB']

    def co(self) -> Vector:
        return self.atoms['C']

    def orientation(self) -> Vector:
        """The orientation vector of a designed residue corresponds to the vector between C_alpha and C_beta atoms.
        :return: the orientation vector in absolute coordinates
        """
        return self.cbeta() - self.calpha()

    def secondary_orientation(self) -> Vector:
        """The secondary orientation vector of a designed residue corresponds to the vector between C_alpha and C_O
        atoms.
        :return: the secondary orientation vector in absolute coordinates
        """
        return self.co() - self.calpha()


def get_icoor_from_aresidue(residue: AbstractResidue) -> InternalCoordinates:
    """Defines an internal coordinate system relative to a given context residue as follows:
    - the origin is the context C_alpha atom
    - the x-axis is defined by the vector between context C_alpha and C_beta.
    - C_alpha, C_beta, and C lie within the xy-plane, such that the internal z axis denotes the normal of this plane.
    - C_N has always a negative z coordinate (when considering L amino acids)
    - lengths are preserved, no scaling is applied when compared to the external coordinate system.
    - C points to positive y direction! -> y is now -y, otherwise D-AAs are created
    :param residue: the context residue of the created internal coordinate system
    :return: a corresponding internal coordinate system
    """
    # basis vectors for the internal coordinate system
    internal_x = (residue.cbeta() - residue.calpha()).normalized()
    planar = (residue.co() - residue.calpha()).normalized()
    internal_z = (internal_x ** planar).normalized()
    internal_y = -(internal_x ** internal_z).normalized()
    # create the rotation matrix
    rotation_matrix: Matrix = [
        [internal_x[0], internal_x[1], internal_x[2]],
        [internal_y[0], internal_y[1], internal_y[2]],
        [internal_z[0], internal_z[1], internal_z[2]]]
    # return the internal coordinates. The translation vector equals the external C_alpha vector.
    return InternalCoordinates(residue.calpha(), rotation_matrix)


class DesignedLigandResidue(AbstractResidue):
    """A designed ligand residue is a residue with an individual position and orientation, which has in addition a
    ligand residue type assigned."""

    def __init__(self, atoms: Dict[str, Vector], chain: str, residue_id: int, restype: str):
        """Creates a designed ligand residue.
        :param atoms: the atoms of the residue and their positions
        :param chain: the identifier of the chain this residue is part of
        :param residue_id: a numerical residue identifier unique within the chain
        :param restype: the amino acid type of this designed ligand residues in three letter code
        """
        super().__init__(atoms)
        self.chain = chain
        self.residue_id = residue_id
        self.restype = restype


class DesignedLigand:
    """instances of this class represent designed ligands as a sequence of consecutive designed residues."""

    def __init__(self, residues: List[DesignedLigandResidue]):
        """Creates a new designed ligand instance.
        :param residues: designed residues comprising the initial contents of the ligand
        """
        self.residues = residues

    def get_title(self) -> str:
        """
        :return: a human readable representation of the ligand sequence.
        """
        title = ""
        for res in self.residues:
            title += seq1(res.restype)
        return title

    def get_min_dim(self) -> Tuple[int, int, int]:
        """
        :return: a tuple containing the minmal x, y, and z coordinates of any residues C_alpha atom
        """
        min_x: int = None
        min_y: int = None
        min_z: int = None
        for res in self.residues:
            if min_x is None or res.calpha()[0] < min_x:
                min_x = res.calpha()[0]
            if min_y is None or res.calpha()[1] < min_y:
                min_y = res.calpha()[1]
            if min_z is None or res.calpha()[2] < min_z:
                min_z = res.calpha()[2]
        return min_x, min_y, min_z

    def get_max_dim(self) -> Tuple[int, int, int]:
        """
        :return: a tuple containing the maximum x, y, and z coordinates of any residues C_alpha atom
        """
        max_x: int = None
        max_y: int = None
        max_z: int = None
        for res in self.residues:
            if max_x is None or res.calpha()[0] > max_x:
                max_x = res.calpha()[0]
            if max_y is None or res.calpha()[1] > max_y:
                max_y = res.calpha()[1]
            if max_z is None or res.calpha()[2] > max_z:
                max_z = res.calpha()[2]
        return max_x, max_y, max_z


def get_designed_ligand_from_scaffold(scaffold: Structure, ligand_info: List[LigandInfo]) -> DesignedLigand:
    """Extracts a designed ligand from a scaffold PDB structure and a list of ligand information structures.
    :param scaffold: a scaffold PDB structure that contains both the binder and the ligand in different chains
    :param ligand_info: describes how the designed ligand is extracted from the scaffold
    :return: a designed ligand constructed from the scaffold and the ligand info
    """
    residues: List[DesignedLigandResidue] = []
    for chain in scaffold.get_chains():
        for (chain_id, residue_id, new_residue_type) in ligand_info:
            if chain.get_id() == chain_id:
                for residue in chain.get_residues():
                    if str(residue.get_id()[1]) == residue_id:
                        if is_amino_acid_residue(residue):
                            atoms: Dict[str, Vector] = {}
                            for atom in residue.get_atoms():
                                atoms[atom.get_name()] = atom.get_vector()
                            if 'CB' not in atoms:
                                atoms['CB'] = get_cbeta_position(residue)
                            residues.append(DesignedLigandResidue(atoms, chain_id, int(residue_id), new_residue_type))
    return DesignedLigand(residues)


class DesignedBinderResidue(AbstractResidue):
    """A designed binder residue is a residue with an individual position and orientation, with no residue type
     information attached."""

    def __init__(self, atoms: Dict[str, Vector], chain: str, residue_id: int, original_residue_type: str):
        """Creates a new designed binder residue from existing data parameters.
        :param atoms: the atoms of the residue and their positions
        :param chain: the identifier of the chain this residue is part of
        :param residue_id: id of the residue in the scaffold's binder chain
        :param original_residue_type: residue type in the scaffold
        """
        super().__init__(atoms)
        self.chain = chain
        self.residue_id = residue_id
        self.original_residue_type = original_residue_type

    def transform_coor(self, icoor: InternalCoordinates, ext_to_int: bool) -> 'DesignedBinderResidue':
        """Converts this binder residue's atom coordinates from external to internal representation or vice versa.
        :param icoor: the coordinate system to use for transformation
        :param ext_to_int: if True, convert external to internal; if False, convert internal to external coordinates.
        :return: a transformed binder residue
        """
        trans_coor: Dict[str, Vector] = {}
        for name, pos in self.atoms.items():
            trans_pos = icoor.external_to_internal(pos, False) if ext_to_int else icoor.internal_to_external(pos, False)
            trans_coor[name] = trans_pos
        return DesignedBinderResidue(trans_coor, self.chain, self.residue_id, self.original_residue_type)


class DesignedBinder:
    """Instances of this class represent designed binders as a collection of (not necessarily connected) residues."""

    def __init__(self, residues: List[DesignedBinderResidue]):
        """Creates a new designed binder instance.
        :param residues initial collection of residues
        """
        self.residues = residues


def get_designed_binder_from_scaffold(scaffold: Structure, binder_info: List[BinderInfo]) -> DesignedBinder:
    """Extracts a designed binder from a scaffold PDB structure and a list of binder information structures.
    Adds a CB atom to the scaffold binder residues, if it is not there!!!

    :param scaffold: a scaffold PDB structure that contains both the binder and the ligand in different chains
    :param binder_info: explains how to extract designed binder residues from the scaffold
    :return: a designed binder with the extracted residues
    """
    residues: List[DesignedBinderResidue] = []
    for chain in scaffold.get_chains():
        for (chain_id, residue_id) in binder_info:
            if chain.get_id() == chain_id:
                for residue in chain.get_residues():
                    if str(residue.get_id()[1]) == residue_id:
                        original_restype = residue.get_resname()

                        # Add a dummy CB atom to the residues, if none is there. (Will show up in the structure!)
                        if 'CB' not in (atm.id for atm in residue.get_atoms()) and residue.resname != "GLY":
                            residue.add(Atom("CB", get_cbeta_position(residue), bfactor=0, occupancy=1.0,
                                             altloc=" ", fullname=f" CB ", serial_number=0, element="C"))
                        if is_amino_acid_residue(residue):
                            atoms: Dict[str, Vector] = {}
                            for atom in residue.get_atoms():
                                atoms[atom.get_name()] = atom.get_vector()
                            if residue.resname == "GLY":  # We need a dummy CB for Glycine
                                atoms["CB"] = get_cbeta_position(residue)
                            residues.append(DesignedBinderResidue(atoms, chain_id, int(residue_id), original_restype))
    return DesignedBinder(residues)



"""This module contains definitions for the data structure 'atlas', a collection of relative positions and orientations
of ligand and binder residues, seen from the subjective perspective of the binder. The module also contains methods for
the extraction of atlas data from PDB files.

:Author: Josef Kynast <josef.kynast@uni-bayreuth.de>
:Author: Felix Schwaegerl <felix.schwaegerl@uni-bayreuth.de>
:date: 2018-02-28
"""
import json
import logging
import os
from collections import OrderedDict, defaultdict
from concurrent.futures import ThreadPoolExecutor
from typing import List, Iterator, Dict, Callable, Any, Union, TextIO

# import h5py
from Bio.PDB.Atom import Atom
from Bio.PDB.Chain import Chain
from Bio.PDB.Model import Model
from Bio.PDB.PDBParser import PDBParser
from Bio.PDB.Residue import Residue
from Bio.PDB.Structure import Structure
from Bio.PDB.vectors import Vector

from atligator.pdb_util import find_ligands, get_icoor_from_pdb_residue, is_amino_acid_residue, get_cbeta_position, \
    get_residues_within_radius, get_path, canonical_amino_acids

logger = logging.getLogger(__name__)


class AtlasDatapoint:
    """An atlas datapoint emerges from an interaction detected between two residues part of a ligand and binder chain,
    respectively."""

    @classmethod
    def from_json(cls, json_datapoint: Dict[str, Any]):
        """
        Return an AtlasDatapoint from a json representation of one.
        :param json_datapoint: json representation of an AtlasDatapoint
        :return: AtlasDatapoint based on json
        """
        return AtlasDatapoint(ligand_restype=json_datapoint['ligand_restype'],
                              binder_restype=json_datapoint['binder_restype'],
                              ligand_origin=json_datapoint['ligand_origin'],
                              binder_origin=json_datapoint['binder_origin'],
                              ligand_atoms=get_atoms_from_json(json_datapoint['ligand_atoms']),
                              binder_atoms=get_atoms_from_json(json_datapoint['binder_atoms']))

    def __init__(self, ligand_restype: str, binder_restype: str, ligand_origin: str or None, binder_origin: str or None,
                 ligand_atoms: Dict[str, Vector], binder_atoms: Dict[str, Vector]):
        """
        Creates a new instance.
        :param ligand_restype: ligand residue type
        :param binder_restype: binder residue type
        :param ligand_origin: path of the ligand residue where the data stems from
        :param binder_origin: path of the binder residue
        :param ligand_atoms: positions of all sidechain atoms of the ligand residue
        :param binder_atoms: positions of all sidechain atoms of the binder residue
        """
        self.ligand_restype = ligand_restype
        self.binder_restype = binder_restype
        self.ligand_origin = ligand_origin
        self.binder_origin = binder_origin
        self.ligand_atoms = ligand_atoms
        self.binder_atoms = binder_atoms

    def binder_calpha(self):
        """
        :return: the CA atom of the binder
        """
        return self.binder_atoms['CA']

    def binder_cbeta(self):
        """
        :return: the CB atom of the binder
        """
        return self.binder_atoms['CB']

    def binder_co(self):
        """
        :return: the C atom of the binder
        """
        return self.binder_atoms['C']

    def binder_orientation(self) -> Vector:
        """
        :return: the vector between CA and CB atoms of the binder
        """
        return self.binder_cbeta() - self.binder_calpha()

    def binder_secondary_orientation(self) -> Vector:
        """
        :return: the vector between CA and C atoms of the binder
        """
        return self.binder_co() - self.binder_calpha()

    def binder_to_pdb_residue(self, res_id: int) -> Residue:
        """Converts the information contained in the binder atoms of this datapoint into a PDB compliant residue
        :param res_id: the residue ID the result will have
        :return: the corresponding PDB residue
        """
        res = Residue(('', res_id, ' '), self.binder_restype, '')
        for atom_name, atom_position in self.binder_atoms.items():
            if self.binder_restype == 'GLY' and atom_name == 'CB':
                continue
            atom = Atom(atom_name, atom_position, 0.0, 1.0, ' ', atom_name, 0, atom_name[0])
            res.add(atom)
        return res

    def ligand_to_pdb_residue(self, res_id: int) -> Residue:
        """Converts the information contained in the ligand atoms of this datapoint into a PDB compliant residue
        :param res_id: the residue ID the result will have
        :return: the corresponding PDB residue
        """
        res = Residue(('', res_id, ' '), self.ligand_restype, '')
        for atom_name, atom_position in self.ligand_atoms.items():
            if self.ligand_restype == 'GLY' and atom_name == 'CB':
                continue
            atom = Atom(atom_name, atom_position, 0.0, 1.0, ' ', atom_name, 0, atom_name[0])
            res.add(atom)
        return res


class Atlas:
    """an atlas is a collection of atlas residues."""

    @classmethod
    def fill_stats(cls, stats):
        def full():
            return {k: 0 for k in canonical_amino_acids}

        full_stats = defaultdict(full, stats)
        for aa in canonical_amino_acids:
            full_stats[aa] = defaultdict(int, full_stats[aa])
            for aa2 in canonical_amino_acids:
                full_stats[aa][aa2] = full_stats[aa][aa2]

        return full_stats

    @classmethod
    def from_json(cls, json_str_or_file):
        """
        Return an Atlas from a json representation of one.
        :param json_str_or_file: json representation of an AtlasDatapoint or open file handle containing it
        :return: Atlas based on json
        """

        def convert(raw_atlas):
            datapoints = []
            for datapoint in raw_atlas['datapoints']:
                datapoints.append(AtlasDatapoint.from_json(datapoint))
            return Atlas(datapoints=datapoints, features=raw_atlas['features'])

        if isinstance(json_str_or_file, str):
            return convert(json.loads(json_str_or_file))
        else:
            return convert(json.load(json_str_or_file))

    def to_json(self):
        """
        Returns the json represenation of an Atlas
        :return: json represenation of an Atlas
        """
        return json.dumps(self, cls=AtlasEncoder)

    def to_json_file(self, fp):
        """
        Dumps the json represenation of an Atlas to an open file handle
        :param fp: open file handle to write json into
        :return: None
        """
        return json.dump(self, fp=fp, cls=AtlasEncoder)

    def __init__(self, datapoints: List[AtlasDatapoint], features: Dict = None):
        """creates a new atlas instance, which is represented explicitly by its datapoints.
        :param datapoints: a sequence of residues represented in the internal coordinate system of the respective
        context ligand.
        """
        self.datapoints = datapoints
        if features is None:
            self.features = {}
        else:
            self.features = features

    def group_by_ligand_origin_dep(self) -> OrderedDict:  # OrderedDict[str, List[AtlasResidue]]
        """Disjointly and completely partitions the data points contained in this atlas by their ligand origin.
        Was discontinued because left behind same-ligand-origin datapoints that are not in a row.
        :return: maps the ligand origin string to the subset of atlas residues it is connected with.
        """
        result: OrderedDict = OrderedDict({})  # OrderedDict[str, List[AtlasResidue]]
        curr_lig_origin = None
        curr_datapoints = []
        for datapoint in self.datapoints:
            if datapoint.ligand_origin != curr_lig_origin:
                if len(curr_datapoints) > 0:
                    result[curr_lig_origin] = curr_datapoints
                    curr_datapoints = []
                curr_lig_origin = datapoint.ligand_origin
            curr_datapoints.append(datapoint)
        if len(curr_datapoints) > 0:
            result[curr_lig_origin] = curr_datapoints
        return result

    def group_by_ligand_origin(self) -> Dict:  # Dict[str, List[AtlasResidue]]
        """Disjointly and completely partitions the data points contained in this atlas by their ligand origin.
        :return: maps the ligand origin string to the subset of atlas residues it is connected with.
        """
        result = defaultdict(list)
        for datapoint in self.datapoints:
            result[datapoint.ligand_origin].append(datapoint)
        return result

    def filter(self, ligand_restype: str = None, binder_restype: str = None) -> 'Atlas':
        """Filters this atlas by a given ligand and binder residue type.
        :param ligand_restype: if None, no filtering by ligand is applied.
        :param binder_restype: if None, no filtering by binder is applied.
        :return: the filtered atlas.
        """
        filtered_dps = self.datapoints[:]
        if ligand_restype is not None:
            filtered_dps = [dp for dp in filtered_dps if dp.ligand_restype == ligand_restype]
        if binder_restype is not None:
            filtered_dps = [dp for dp in filtered_dps if dp.binder_restype == binder_restype]
        return Atlas(filtered_dps)

    def get_stats(self, ligand_per_binder: bool = False) -> Dict[str, Dict[str, int]]:
        """Extracts statistical information about the distribution of datapoints from this atlas.
        :param ligand_per_binder: if True, the dictionary maps ligand to binder, else vice versa.
        :return: a dictionary that indicates the number of data points for every ligand (first key) and binder
        (second key) residue type.
        """
        stats: Dict[str, Dict[str, int]] = {}
        for dp in self.datapoints:
            key1 = dp.binder_restype if ligand_per_binder else dp.ligand_restype
            key2 = dp.ligand_restype if ligand_per_binder else dp.binder_restype
            if key1 not in stats:
                stats[key1] = {}
            if key2 not in stats[key1]:
                stats[key1][key2] = 0
            stats[key1][key2] = stats[key1][key2] + 1
        return stats

    def to_pdb_structure(self, structure_name: str) -> Structure:
        """
        :return: A PDB compliant structure that contains the datapoints of this atlas. By convention, chain A contains
        different conformations of the ligand residues of all datapoints, chains B contains binder residues.
        """
        model = Model(0)
        resid = 1
        ligand_chain = Chain('A')
        binder_chain = Chain('B')
        for dp in self.datapoints:
            ligres = dp.ligand_to_pdb_residue(resid)
            ligand_chain.add(ligres)
            resid += 1
        for dp in self.datapoints:
            binres = dp.binder_to_pdb_residue(resid)
            binder_chain.add(binres)
            resid += 1
        model.add(ligand_chain)
        model.add(binder_chain)
        struc = Structure(structure_name)
        struc.add(model)
        return struc

    def __len__(self):
        return len(self.datapoints)


def find_datapoints(model: Model, ir_default: float = 4.0, ir_hbond: float = 6.0, ir_aromatic: float = 6.0,
                    ir_ionic: float = 8.0, minlen: int = 4, maxlen: int = 20, include_hydrogens: bool = False,
                    alternative_lig_aa: str or None = None, restrict_to_alternative: bool = True,
                    allow_self_interactions: bool = False, skip_bb_atoms: bool = False) -> Iterator[AtlasDatapoint]:
    """Processes a PDB structure, identifies ligand chains in there, looks for interactions between ligand residues and
    binder residues. Converts each interacting binder residue into an atlas residue in internal coordinates, and yields
    all of these as results.
    :param model: the PDB model to process
    :param ir_default: interaction radius for interaction between arbitrary atoms in Angstrom
    :param ir_hbond: interaction radius for hydrogen bond interactions in Angstrom
    :param ir_aromatic: interaction radius for aromatic interactions in Angstrom
    :param ir_ionic: interaction radius for ionic interactions in Angstrom
    :param minlen: the minimum length of a chain in order to be considered as a ligand
    :param maxlen: the maximum length of a chain in order to be considered as a ligand
    :param include_hydrogens: whether hydrogen atoms shall be considered for interaction identification
    :param alternative_lig_aa: If given this alternative name of non-canonical aas will be also searched for as a
    ligand residue
    :param restrict_to_alternative: If True: If alternative Ligand amino acid is given restrict to this type.
    :param allow_self_interactions: If True interactions within one chain are taken into account (Slows down
    calculation heavily!!)
    :return: a generator of AtlasDatapoint instances
    """
    ligands = find_ligands(model, minlen, maxlen, alternative_lig_aa)
    for ligand in ligands:
        for ligand_residue in ligand.get_residues():
            # If alternative Ligand amino acid is given restrict to this type.
            if alternative_lig_aa is not None and \
                    restrict_to_alternative and not ligand_residue.get_resname() == alternative_lig_aa:
                continue
            if not (is_amino_acid_residue(ligand_residue, extended=alternative_lig_aa)):
                continue
            ligres: str = ligand_residue.get_resname()
            icoor = get_icoor_from_pdb_residue(ligand_residue)
            for binder_residue in get_residues_within_radius(model, ligand, ligand_residue, ir_default,
                                                             ir_hbond, ir_aromatic, ir_ionic,
                                                             include_hydrogens=include_hydrogens,
                                                             allow_ligand_residues=allow_self_interactions,
                                                             skip_bb_atoms=skip_bb_atoms):
                if not (is_amino_acid_residue(binder_residue)):
                    continue
                ligand_atoms: Dict[str, Vector] = {}
                for ligand_atom in ligand_residue.get_atoms():
                    if not ligand_atom.element == 'H':
                        ligand_atoms[ligand_atom.get_id()] = icoor.external_to_internal(ligand_atom.get_vector(), False)
                if 'CB' not in ligand_atoms:
                    ligand_atoms['CB'] = icoor.external_to_internal(get_cbeta_position(ligand_residue), False)
                binder_atoms: Dict[str, Vector] = {}
                for binder_atom in binder_residue.get_atoms():
                    if not binder_atom.element == 'H':
                        binder_atoms[binder_atom.get_id()] = icoor.external_to_internal(binder_atom.get_vector(), False)
                if 'CB' not in binder_atoms:
                    binder_atoms['CB'] = icoor.external_to_internal(get_cbeta_position(binder_residue), False)
                ligand_origin = get_path(ligand_residue)
                binder_origin = get_path(binder_residue)
                yield AtlasDatapoint(ligres, binder_residue.get_resname(), ligand_origin, binder_origin,
                                     ligand_atoms, binder_atoms)


def generate_atlas(filenames: List[str], ir_default: float = 4.0, ir_hbond: float = 6.0, ir_aromatic: float = 6.0,
                   ir_ionic: float = 8.0, min_ligand_len: int = 3, max_ligand_len: int = 20, n_workers: int = -1,
                   include_hydrogens: bool = False, include_alternative_models: bool = False,
                   observer: Callable[[str], None] = None, alternative_lig_aa: str = None,
                   restrict_to_alternative: bool = True, allow_self_interactions: bool = False,
                   skip_bb_atoms: bool = True) -> Atlas:
    """Generates an atlas from a list of PDB files. In a multithreaded way, datapoints are collected and combined into
    a single atlas instance, which is returned as result.
    :param filenames: a list of PDB file names to process
    :param ir_default: interaction radius for interaction between arbitrary atoms in Angstrom
    :param ir_hbond: interaction radius for hydrogen bond interactions in Angstrom
    :param ir_aromatic: interaction radius for aromatic interactions in Angstrom
    :param ir_ionic: interaction radius for ionic interactions in Angstrom
    :param min_ligand_len: minimum chain length. see find_datapoints
    :param max_ligand_len: maximum chain length. see find_datapoints
    :param n_workers: size of the threadpool, or -1 if it shall equal the number of physical CPU cores
    :param include_hydrogens: whether hydrogen atoms are to be considered. see find_datapoints
    :param include_alternative_models: whether to include data from models different from the first model of the
    structure (i.e., whether not to ignore alternative NMR conformations)
    :param observer: a callback function that receives the file name of every processed input structure
    :param alternative_lig_aa: If given this alternative name of non-canonical aas will be also searched for as a
    ligand residue
    :param restrict_to_alternative: If True: If alternative Ligand amino acid is given restrict to this type.
    :param allow_self_interactions: If True interactions within one chain are taken into account (Slows down
    calculation heavily!!)
    :param skip_bb_atoms: If True, interactions with the ligand backbone atoms are not considered.
    :return: The atlas created from the PDB files based on the given parameters
    """
    if n_workers == -1:
        n_workers = os.cpu_count()

    def __open_and_add_datapoints(_filename: str) -> List[AtlasDatapoint]:
        """Worker function that creates a PDB parser, opens a PDB file, calls find_datapoints, and returns the result.
        :param _filename: the PDB file to process
        :return: a list of datapoints that can be appended to an atlas
        """
        _parser = PDBParser(QUIET=True)
        # TODO pathlib
        _structure: Structure = _parser.get_structure(_filename.split('/')[-1], _filename)
        _datapoints: List[AtlasDatapoint] = []
        for _m in _structure:
            _datapoints.extend(find_datapoints(model=_m, ir_default=ir_default, ir_hbond=ir_hbond,
                                               ir_aromatic=ir_aromatic, ir_ionic=ir_ionic, minlen=min_ligand_len,
                                               maxlen=max_ligand_len,
                                               include_hydrogens=include_hydrogens,
                                               alternative_lig_aa=alternative_lig_aa,
                                               restrict_to_alternative=restrict_to_alternative,
                                               allow_self_interactions=allow_self_interactions,
                                               skip_bb_atoms=skip_bb_atoms
                                               ))
            if not include_alternative_models:
                break
        if observer is not None:
            observer(_filename)
        return _datapoints

    atlas = Atlas([], features={"ir_default": ir_default, "ir_hbond": ir_hbond, "ir_aromatic": ir_aromatic,
                                "ir_ionic": ir_ionic, "minlen": min_ligand_len, "maxlen": max_ligand_len,
                                "include_hydrogens": include_hydrogens,
                                "include_alternative_models": include_alternative_models,
                                "alternative_lig_aa": alternative_lig_aa,
                                "restrict_to_alternative": restrict_to_alternative,
                                "allow_self_interactions": allow_self_interactions, "skip_bb_atoms": skip_bb_atoms})
    with ThreadPoolExecutor(n_workers) as executor:
        for dps in executor.map(__open_and_add_datapoints, filenames):
            atlas.datapoints.extend(dps)
    return atlas


class AtlasGeneration:
    """
    Keeps track of atlas generation and processed input files with a simple observer method.
    """

    def __init__(self, files: List[str]):
        self.files = files
        self.processed = set()
        self.total = len(files)

    def simple_observer(self, file):
        self.processed.add(file)
        logger.info(f"Finished processing with {len(self.processed)} of {self.total} input structures.\t({file})")

    def generate_atlas(self, ir_default: float = 4.0, ir_hbond: float = 6.0, ir_aromatic: float = 6.0,
                       ir_ionic: float = 8.0, min_ligand_len: int = 3, max_ligand_len: int = 20, n_workers: int = -1,
                       include_hydrogens: bool = False, include_alternative_models: bool = False,
                       observer: Callable[[str], None] = None, alternative_lig_aa: str = None,
                       restrict_to_alternative: bool = True, allow_self_interactions: bool = False,
                       skip_bb_atoms: bool = True) -> Atlas:
        return generate_atlas(filenames=self.files,
                              ir_default=ir_default,
                              ir_hbond=ir_hbond,
                              ir_aromatic=ir_aromatic,
                              ir_ionic=ir_ionic,
                              min_ligand_len=min_ligand_len,
                              max_ligand_len=max_ligand_len,
                              n_workers=n_workers,
                              include_hydrogens=include_hydrogens,
                              include_alternative_models=include_alternative_models,
                              alternative_lig_aa=alternative_lig_aa,
                              restrict_to_alternative=restrict_to_alternative,
                              allow_self_interactions=allow_self_interactions,
                              observer=self.simple_observer if observer is None else observer,
                              skip_bb_atoms=skip_bb_atoms)


def atlas_to_json_file(atlas: Atlas, fp):
    """
    Dumps the json represenation of an Atlas to an open file handle
    :param atlas: Atlas to json convert
    :param fp: open file handle to write json into
    :return: None
    """
    return atlas.to_json_file(fp=fp)


def atlas_to_json(atlas: Atlas):
    """
    Returns the json represenation of an Atlas
    :param atlas: Atlas to json convert
    :return: json represenation of an Atlas
    """
    return atlas.to_json()


def json_to_atlas(json_str_or_file: Union[str, TextIO]):
    """
    Return an Atlas from a json representation of one.
    :param json_str_or_file: json representation of an Atlas or open file handle containing it
    :return: Atlas based on json
    """
    return Atlas.from_json(json_str_or_file=json_str_or_file)


def get_atoms_from_json(atom_dict: Dict[str, Dict]):
    """
    Return an atom representation as a Dict of Vectors from a json representation of one.
    :param atom_dict: json representation of atoms
    :return: dictionary with atom names and Vectors based on json
    """
    return {atom: Vector(**vector) for atom, vector in atom_dict.items()}


class AtlasEncoder(json.JSONEncoder):
    """
    Enables the json Encoder to also encode an Atlas or AtlasDatapoint or Vector
    """

    def default(self, obj):
        if isinstance(obj, Vector):
            return {'x': obj[0],
                    'y': obj[1],
                    'z': obj[2]}
        if isinstance(obj, AtlasDatapoint):
            return {'ligand_restype': obj.ligand_restype,
                    'binder_restype': obj.binder_restype,
                    'ligand_origin': obj.ligand_origin,
                    'binder_origin': obj.binder_origin,
                    'ligand_atoms': obj.ligand_atoms,
                    'binder_atoms': obj.binder_atoms}
        elif isinstance(obj, Atlas):
            return {'features': obj.features,
                    'datapoints': obj.datapoints}
        else:
            return json.JSONEncoder.default(self, obj)



"""A ligand-binder complex consists of a designed ligand, a designed binder, and a sequence of aligned residues which
originate from an atlas. This module contains data structures and operations for creating and manipulating such
ligand-binder complexes.

:Author: Josef Kynast <josef.kynast@uni-bayreuth.de>
:Author: Felix Schwaegerl <felix.schwaegerl@uni-bayreuth.de>
:date: 2018-03-05
"""

from typing import List, Dict, Any

from Bio.PDB.Structure import Structure
from Bio.PDB.vectors import Vector

from atligator.atlas import Atlas
from atligator.structure import AbstractResidue, DesignedLigand, DesignedBinder, LigandInfo, BinderInfo, CouplingInfo, \
    get_designed_binder_from_scaffold, get_designed_ligand_from_scaffold, get_icoor_from_aresidue


class ComplexDescriptor:
    """A complex descriptor complements a scaffold PDB file with additional information necessary for the construction
    of a designed ligand and binder out of the scaffold."""

    def __init__(self, ligand_info: List[LigandInfo], binder_info: List[BinderInfo], coupling_info: List[CouplingInfo]):
        """Creates a new complex descriptor instance from the data it is going to contain.
        :param ligand_info: a list of tuples, each containing chain id, residue id, and designed residue type
        :param binder_info: a list of tuples, each containing chain id and residue id
        :param coupling_info: a list of coupling information. The first element denotes the chain id, the second element
        is a list of coupled residues.
        """
        self.ligand_info = ligand_info
        self.binder_info = binder_info
        self.coupling_info = coupling_info


def read_complex_descriptor(filename: Any) -> ComplexDescriptor:
    """Reads in a descriptor file and returns a corresponding instance of ComplexDescriptor.
    :param filename: Pathname of a file to be opened and read or File itself
    :return a corresponding instance of ComplexDescriptor
    """

    def _get_properties(file):
        linfo: List[LigandInfo] = []
        binfo: List[BinderInfo] = []
        cinfo: List[CouplingInfo] = []
        stage = 0
        for line in file:
            if len(line.split()) <= 1:
                stage += 1
                continue
            content = line.split()
            if stage == 0:
                linfo.append((content[0], content[1], content[2]))
            elif stage == 1:
                binfo.append((content[0], content[1]))
            elif stage == 2:
                cinfo.append((content[0], content[1:]))
        return linfo, binfo, cinfo

    if not isinstance(filename, str):
        ligand_info, binder_info, coupling_info = _get_properties(filename)
    else:
        with open(filename, 'r') as fo:
            ligand_info, binder_info, coupling_info = _get_properties(fo)
    return ComplexDescriptor(ligand_info, binder_info, coupling_info)


class AlignedResidue(AbstractResidue):
    """An aligned residue is a residue with an individual position and orientation, which has in addition a binder
    residue type as well as a position of the originating residue in the designed ligand assigned."""

    def __init__(self, atoms: Dict[str, Vector], binder_restype: str, ligand_res_position: int, size_factor: int):
        """Creates a new instance.
        :param atoms: atoms part of this residues and their positions
        :param binder_restype: binder residue type inferred from the atlas
        :param ligand_res_position: the residue id relative to the ligand chain
        :param size_factor: number of datapoints of the original atlas
        """
        super().__init__(atoms)
        self.binder_restype = binder_restype
        self.ligand_res_position = ligand_res_position
        self.size_factor = size_factor


class LigandBinderComplex:
    """A ligand-binder complex is created from a scaffold structure. It consists of a designed ligand structure, a set
    of designed binder residues, and a set of aligned residues inferred from the atlas."""

    def __init__(self, ligand: DesignedLigand, binder: DesignedBinder, aligned_residues: List[AlignedResidue]):
        """Creates a new instance. All vectors are supposed to be in absolute coordinates (defined by the scaffold)
        :param ligand: the designed ligand part of the complex
        :param binder: the designed binder part of the complex
        :param aligned_residues: a set of aligned residue
        """
        self.ligand = ligand
        self.binder = binder
        self.aligned_residues = aligned_residues


def generate_complex(atlas: Atlas or None, scaffold: Structure, descriptor: ComplexDescriptor) -> LigandBinderComplex:
    """Generates a ligand-binder complex out of an atlas database, a scaffold structure, and a descriptor that refers
    to the scaffold structure.
    :param atlas: the atlas from which aligned residues will be inferred from
    :param scaffold: a scaffold PDB structure from which the position and orientation of residues of both the designed
    binder and the designed ligand will be extracted.
    :param descriptor: a descriptor file that contains information about ligand and binder residues and types
    :return: the generated ligand-binder complex
    """
    ligand = get_designed_ligand_from_scaffold(scaffold, descriptor.ligand_info)
    binder = get_designed_binder_from_scaffold(scaffold, descriptor.binder_info)
    aligned_residues: List[AlignedResidue] = []
    if atlas is not None:
        for lresid, ligres in enumerate(ligand.residues):
            icoor = get_icoor_from_aresidue(ligres)
            relevant_datapoints = [d for d in atlas.datapoints if d.ligand_restype == ligres.restype]
            size_factor = len(relevant_datapoints)
            for datapoint in relevant_datapoints:
                atoms: Dict[str, Vector] = {}
                for atom, int_coor in datapoint.binder_atoms.items():
                    ext_coor = icoor.internal_to_external(int_coor, False)
                    atoms[atom] = ext_coor
                ares = AlignedResidue(atoms, datapoint.binder_restype, lresid, size_factor)
                aligned_residues.append(ares)
    return LigandBinderComplex(ligand, binder, aligned_residues)



import io
import pickle
import sys
from argparse import FileType, Namespace
from tkinter import font, Tk, Button, ttk, Label, StringVar, BooleanVar, filedialog, Checkbutton, Radiobutton, IntVar, \
    Spinbox, Frame, W, messagebox, Toplevel
from typing import Dict, Any, List, Tuple

from atligator.acomplex import read_complex_descriptor, ComplexDescriptor
from atligator.argument_collection import help_txts, arg
from atligator.pdb_util import canonical_amino_acids as aa


def next_row(current: int, value: int = None):
    """
    Returns an incremented version of current value. if value is set the sum of current and value is returned.
    Is used for row numbers for grid window manager.
    :param current: Is returned. Either +1 or +value if value is stated.
    :param value: Added to current if stated
    :return: current + 1 or current + value (if stated)
    """
    if value is None:
        return 1 + current
    else:
        return current + value


def normalize_couples(couples: List[List[str]]) -> List[List[str]]:
    """
    Will normalize a coupling list. This list consists of lists with strings. These inner lists represent the rows
    in coupling section of the descriptor. Thus, all residues represented by strings in one row are coupled in mutation.
    Normalization will add empty strings ("") in these row lists so the residue numbers increase in the right order:
    Columns, then rows. For example:
    1 5          1 ''  11
    2 6         ''  6  13
    3 7   or     3 ''  15        are saved as [['1', '5'],['2', '6'],['3', '7'],['4', '8']]
    4 8          4  8  ''                  or [['1', '', '11'],['', '6', '13'],['3', '', '15'],['4', '8', '']]
    :param couples: List of list with strings representing residues or wildcards ('')
    :return: Same list with inserted wildcards to follow the right order (increasing in columns then rows)
    """

    def sort_row(list_list: List[List[str]]) -> List[List[str]]:
        """
        Sorts rows of a matrix (list of list with int as strings) and returns it with strings again
        :param list_list: list of lists to sort
        :return: sorted lists in list
        """
        for r_i, r in enumerate(list_list):
            for x in range(len(r)):
                r[x] = int(r[x])
            list_list[r_i] = sorted(r)
        for r in list_list:
            for y in range(len(r)):
                r[y] = str(r[y])
        return list_list

    def next_non_zero(list_list: List[List[str]], col: int, row_i: int, x_y: str = "x") -> int or float:
        """
        Returns the next residue id that is not zero. If the next one is zero it walks in a certain direction (x_y)
        and returns the next non zero value or an incredibly high value if it doesn't find any.
        :param list_list: List of lists to look for this residue id
        :param col: starting point: column number
        :param row_i: starting point: row number
        :param x_y: direction to walk
        :return: int value of residue id
        """
        value = 0
        j = 0
        while value == 0:
            if x_y == "x":
                value = int(list_list[row_i][col + j])
            elif x_y == "y":
                if row_i + j == len(list_list) or col >= len(list_list[row_i + j]):
                    return j, 10000
                value = int(list_list[row_i + j][col])
            else:
                sys.exit()
            j += 1
        return j - 1, value

    couples = sort_row(couples)
    if all(e == "" for e in couples):
        return [["", ""]]
    not_finished = True
    max_len = 0
    while not_finished:
        not_finished = False
        max_len = max(len(l) for l in couples)
        for u in range(max_len):
            sth_happens = True
            while sth_happens:
                sth_happens = False
                for i, row in enumerate(couples):
                    if i != len(couples) - 1 and u < len(row):
                        c, this_res = next_non_zero(couples, u, i)
                        c, next_res = next_non_zero(couples, u + c, i + 1, "y")
                        if next_res < this_res:
                            row.insert(u, "0")
                            sth_happens = True
                            if len(row) > max_len:
                                not_finished = True
                    elif u < len(row):
                        c, this_res = next_non_zero(couples, u, i)
                        c, next_res = next_non_zero(couples, u + 1 + c, 0, "y")
                        if next_res < this_res:
                            row.insert(u, "0")
                            sth_happens = True
                            if len(row) > max_len:
                                not_finished = True
    for row in couples:
        for _ in range(max_len - len(row)):
            row.append("")
        for res in range(len(row)):
            if row[res] == '0':
                row[res] = ''
    return couples


class Interface:
    def __init__(self, title, main: bool = True):
        if not main:
            self.window = Toplevel()
        else:
            self.window = Tk()
        default_font = font.nametofont("TkDefaultFont")
        default_font.configure(size=12, weight='bold')
        self.bold = font.Font(weight='bold')
        self.window.option_add("*Font", default_font)
        # self.window.configure(background='#e0ffec')
        self.window.title(title)
        # centers the self.window
        # self.window.eval('tk::Place.self.window %s center' % self.window.winfo_pathname(self.window.winfo_id()))
        # self.window.eval('tk::PlaceWindow %s center' % self.window.winfo_toplevel())
        # self.window.geometry('350x200')
        self.row_ = 0

    def run_mode(self):
        pass

    def build_main_window(self, **kwargs):
        pass

    def send_cmd(self):
        pass

    @staticmethod
    def return_default(_) -> Any:
        return None

    def exit_ui(self):
        if isinstance(self, ParameterManager):
            del self.window
            sys.exit()
        else:
            self.window.destroy()

    def insert_sendcancel(self, frame: Frame or Tk or Toplevel = None):
        if frame is None:
            frame = self.window
        stop = Button(frame, text="Cancel", command=self.exit_ui)
        stop.pack(side="right")
        send = Button(frame, text="Send", command=self.send_cmd)
        send.pack(side="right")

    def insert_widget_basics(self, name: str, frame_id: Frame):
        """
        Inserts default widgets for elements such as a help button and a label.
        :param self: ParameterManager object we're working in
        :param name: Name of the element (short version)
        :param frame_id: Frame to put widgets in
        """
        self.insert_help_button(name, frame_id)
        self.insert_element_label(name, frame_id)

    def insert_help_button(self, name: str, frame_id: Frame):
        """
        Inserts a help button for specified element (name) in frame (frame_id).
        :param self: ParameterManager object we're working in
        :param name: Name of the element (short version)
        :param frame_id: Frame to put widgets in
        """
        info_button = Button(frame_id, text="?", fg='white', bg='green', font=self.bold,
                             command=lambda: self.show_info(name))
        info_button.grid(column=0, row=self.row_, sticky=W)

    def insert_element_label(self, name: str, frame_id: Frame or Tk, ret: bool = False,
                             text: str = False, column=1, row=None, sticky=W, fg="black", pack: bool = False,
                             side: str = "right"):
        """
        Inserts a label for specified element (name) in frame (frame_id).
        Special cases (for filebox content or major frames) can be inserted by specifying
        an 'ret' which returns an element that should be assigned to that label
        (For example a dynamic file box content text)

        :param self: ParameterManager object we're working in
        :param name: Name of the element (short version)
        :param frame_id: Frame to put widgets in
        :param ret: Return an element to assign the Label to. If True, text and row needs to be specified, too.
        :param text: Text to insert in the Label. Only taken into account if 'ret' is true
        :param column: Column to (grid) pack the label in
        :param row: Row to (grid) pack the label in
        :param sticky: Adjusts the sticky side. If None, sticky is not enabled.
        :param fg: Adjusts the font color of the label. Only taken into account if 'ret' is true.
        :param pack: Switches to pack window manager instead of grid if enabled
        :param side: Side where to pack label.
        """
        row_ = self.row_ if row is None else row
        if ret is False:
            label_ = Label(frame_id, text="    " + name + ": ")
            if not pack:
                label_.grid(column=column, row=row_, sticky=W)
            else:
                label_.pack(side=side, anchor=sticky)
        else:
            element = Label(frame_id, text=text, fg=fg)
            if not pack:
                element.grid(column=column, row=row_, sticky=sticky)
            else:
                element.pack(side=side, anchor=sticky)
            return element

    def show_info(self, name: str):
        """
        Shows a messagebox with long id of 'name' and its help text
        :param name: name of parameter (e.g. atlas)
        """
        messagebox.showinfo('Default Info', "We all wanna change the world.")


class ParameterManager(Interface):
    def __init__(self, defaults):
        super().__init__("Welcome to atligator!")
        self.window.eval('tk::PlaceWindow %s center' % self.window.winfo_toplevel())

        self.conversion_dict: Dict[str, str] = {}
        for a in arg:
            if 'l_name' in arg[a]:
                self.conversion_dict[arg[a]['l_name']] = a
            else:
                self.conversion_dict[a] = a
        del self.conversion_dict['graphical_input']
        del self.conversion_dict['atpar']

        # Read default parameters from argument parser input
        self.defaults: Dict[str, Any] = {}
        for args in vars(defaults):
            if args not in ("graphical_input", "", "atpar"):
                self.defaults[self.conversion_dict[args]] = vars(defaults)[args]

        # If parameter is mandatory use current value as default:
        for defs in self.defaults:
            if arg[defs]["group"] == "None":
                arg[defs]["default"] = defs

        # Read descriptor as well as ligand length and residues from descriptor
        self.descriptor = read_complex_descriptor(self.defaults['descriptor'].name)
        self.lig_length = len(self.descriptor.ligand_info)
        self.residues: Dict[int, ttk.Combobox] = {i: None for i in range(self.lig_length)}
        self.coupling_info: List[Tuple[str, List[str]]] = []

        # Define groups and save status: if group frame was opened or not
        self.g_checked: Dict[str, Label] = {'mp': None, 'aa': None, 'lc': None,
                                            'fit': None, 'mut': None, 'sc': None,
                                            'cs': None, 'da': None, 'rr': None,
                                            'os': None, 'out': None}

        # Dictionary with major frames for groups
        self.maj_frames: Dict[str, List[Frame]] = {}
        # Dict for checkbutton for opening group frame. Contains label, button and current state
        self.chb: Dict[str, List[Label or Button or BooleanVar]] = {}
        for a in self.g_checked:
            self.maj_frames[a] = [None, None, None]
            self.chb[a] = [None, None, None]

        self.button_dict = {}
        self.spins: Dict[str, Spinbox or None] = {}
        self.checks: Dict[str, Checkbutton or None] = {}
        self.checks_val: Dict[str, BooleanVar or None] = {}
        self.radios_val: Dict[str, StringVar or str] = {}
        self.radios_choices: Dict[str, List[str]] = {}
        self.radios: Dict[str, List[Radiobutton or None]] = {}
        self.radio_frames: Dict[str, Frame] = {}
        self.entrys: Dict[str, Any or None] = {}
        self.files: Dict[str, Any or None] = {}
        self.files_curr: Dict[str, str or None] = {}
        for a in arg:
            if arg[a]['type'] in ('int', 'float'):
                self.spins[a] = None
            if a != "G" and arg[a]['type'] == 'bool':
                self.checks[a] = None
                self.checks_val[a] = None
            if arg[a]['type'] == 'str' and arg[a]['choices'] != "":
                self.radios_val[a] = ""
                self.radios_choices[a] = arg[a]['choices']
                self.radios[a] = [None] * len(arg[a]['choices'])
                self.radio_frames[a] = None
            if arg[a]['type'] == 'str' and arg[a]['choices'] == "" and a != "A":
                self.entrys[a] = None
            if arg[a]['type'] in ("FileType('r')", "FileType('w')") and a != "descriptor":
                self.files[a] = None
                self.files_curr[a] = None
        self.result = {}
        self.namespace = Namespace()
        self.bottom_label = Label(self.window, text="")
        self.bottom_label.pack(side="bottom")
        self.row_ = 0
        self.atpar = None
        for args in vars(defaults):
            if args == "atpar":
                if vars(defaults)[args] is not None:
                    self.atpar = vars(defaults)[args]
                break

    def run_mode(self):
        self.build_main_window()
        # print(self.namespace)
        return self.namespace

    def combine_results(self):

        def modify_descriptor(name: str = "descriptor"):
            def delete_old_coupling(l: List) -> List:
                l.reverse()
                splitter = 0
                for x in l:
                    if x.startswith("-"):
                        splitter += 1
                if splitter < 2:
                    l.insert(0, "-----------\n")
                else:
                    counter = 0
                    for x in l:
                        if x.startswith("-"):
                            break
                        counter += 1
                    if counter > 0:
                        del l[:counter]
                l.reverse()
                return l

            modify_lig_res: bool = not all(self.residues[res].get() == self.descriptor.ligand_info[res][2] for res in
                                           range(self.lig_length))
            modify_coupling: bool = len(self.coupling_info) != 0
            if modify_lig_res or modify_coupling:
                line_list = []
                with open(self.defaults[name].name, "r") as df:
                    for lines in df:
                        line_list.append(lines)
                if modify_lig_res:
                    for j in range(self.lig_length):
                        for char in range(len(line_list[j])):
                            if line_list[j][char:char + 3] in aa:
                                line_list[j] = line_list[j].replace(line_list[j][char:char + 3], self.residues[j].get())
                                break
                if modify_coupling:
                    line_list = delete_old_coupling(line_list)
                    for row in self.coupling_info:
                        for r, res in enumerate(row[1]):
                            row[1][r] = (3 - len(str(res))) * " " + str(res)
                        line_list.append(row[0] + " " + " ".join(row[1]) + "\n")
                with open(self.defaults[name].name, "w") as df:
                    df.write("".join(line_list))
                print("descriptor should have been changed")
                desc_file_func = FileType('r') if arg[name]['type'] == "FileType('r')" else FileType('w')
                self.result[name] = desc_file_func(self.defaults[name].name)
            else:
                self.result[name] = self.defaults[name]
                print("descriptor not changed")

        for i in self.defaults:
            if i == "descriptor":
                # If nothing changed in ligand residues or coupling - take the same as default
                modify_descriptor()

        for i in self.checks_val:  # Include stati of checkboxes to results
            self.result[i] = self.checks_val[i].get()
        for i in self.spins:  # Include all spinbox values to results and convert in int or float
            if "." in self.spins[i].get():
                self.result[i] = float(self.spins[i].get())
            else:
                self.result[i] = int(self.spins[i].get())
        for i in self.radios_val:  # Include the current stati of all radio values to results
            self.result[i] = self.radios_val[i].get()
        for i in self.entrys:  # Include all strings (from Comboboxes or Entrys) into results
            if self.entrys[i].get() == "value unset":
                self.result[i] = None
            else:
                self.result[i] = self.entrys[i].get()
        for i in self.files:  # Include all FileType(s) (from Comboboxes or Entrys or Paths) into results
            if self.files[i] is None:
                self.result[i] = None
            elif i is not "descriptor":
                file_func = FileType('r') if arg[i]['type'] == "FileType('r')" else FileType('w')
                self.result[i] = file_func(self.files[i])

    def send_cmd(self):
        if self.perfect():
            print("sending..")
            self.combine_results()
            for i in self.result:  # Create an argparser Namespace object with all arguments (long identifiers) + values
                long_id: str = list(self.conversion_dict.keys())[list(self.conversion_dict.values()).index(i)]
                setattr(self.namespace, long_id, self.result[i])
            self.window.withdraw()
            self.window.destroy()
        else:
            print("An error was detected. Check ligand residues or coupling.")

    def perfect(self) -> bool:
        if all(self.residues[x].get() in aa for x in self.residues):
            return True
        return False

    def save_params(self):
        """
        Saves all parameters to a pickle file in '*.atpar' format.
        All special formats (incompatible with pickle) are converted first and
        reconverted after importing (import_params).
        """
        def convert(lit: Dict):
            for i in lit:
                if isinstance(lit[i], BooleanVar):
                    lit[i] = bool(lit[i].get())
                elif isinstance(lit[i], StringVar):
                    lit[i] = lit[i].get()
                elif isinstance(lit[i], IntVar):
                    lit[i] = int(lit[i])
                elif isinstance(lit[i], Spinbox):
                    lit[i] = lit[i].get()
                elif isinstance(lit[i], ttk.Combobox):
                    lit[i] = lit[i].get()

        a = dict(self.defaults)
        b = dict(self.checks_val)
        c = dict(self.spins)
        d = dict(self.radios_val)
        e = dict(self.entrys)
        f = dict(self.files)
        g = dict(self.residues)
        h = self.coupling_info
        for part in a:
            if isinstance(a[part], io.TextIOWrapper):
                a[part] = "IOWrapper " + a[part].name
        data: List[Any] = [a, b, c, d, e, f, g]
        for lit_ in data:
            convert(lit_)
        data.append(h)
        par_file = filedialog.asksaveasfilename(filetypes=(("atpar files", "*.atpar"), ("all files", "*.*")))
        if not par_file:
            print("Saving process stopped. Enter valid name")
            self.bottom_label.configure(text="Saving process stopped. Enter valid name!",
                                        fg="red", bg="grey")
            return 0
        with open(par_file, 'wb') as wb:
            pickle.dump(data, wb)
        print("saved in", par_file)
        self.bottom_label.configure(text="Successfully saved to " + par_file + "!", fg="green", bg="grey")

    def import_params(self, init: str or bool = False):
        """
        Imports parameters from a '*.atpar' file (created with save_params).
        Reconverts special types from pickle (e.g. FileType)
        """
        if init:
            par_file = init
        else:
            par_file = filedialog.askopenfilename(filetypes=(("atpar files", "*.atpar"), ("all files", "*.*")))
        if not par_file:
            print("Import process stopped. Enter valid parameter filename")
            self.bottom_label.configure(text="Import process stopped. Enter valid parameter filename!",
                                        fg="red", bg="grey")
            return 0
        with open(par_file, 'rb') as rb:
            try:
                data2 = pickle.load(rb)
            except EOFError:
                print("Import process stopped. Enter valid parameter filename")
                self.bottom_label.configure(text="Import process stopped. Enter valid parameter filename!",
                                            fg="red", bg="grey")
                return 0
        self.defaults = data2[0]
        for check in self.checks_val:
            self.checks_val[check].set(data2[1][check])
        for spin in self.spins:
            self.spins[spin].delete(0, "end")
            self.spins[spin].insert(0, data2[2][spin])
        for radio in self.radios_val:
            self.radios_val[radio].set(data2[3][radio])
        for entry in self.entrys:
            self.entrys[entry].set(data2[4][entry])
        for file in self.files:
            self.files[file] = data2[5][file]
            if self.files_curr[file] is not None:
                self.files_curr[file].configure(text=data2[5][file])
        for part in self.defaults:
            val = self.defaults[part]
            if type(val) == str and val.split()[0] == "IOWrapper":
                file_func = FileType('r') if arg[part]['type'] == "FileType('r')" else FileType('w')
                self.defaults[part] = file_func(val.split()[1])
        for res in self.residues:
            self.residues[res].current(aa.index(data2[6][res]))
        self.coupling_info = data2[7]
        print("imported from", par_file)
        self.bottom_label.configure(text=f"Successfully imported from {par_file}!", fg="green")

    def update_dict(self):
        """
        Updates the checkbuttons for hiding or showing group specific frames.
        Used if a state is changed ('show' checkbutton is (de)activated).
        """
        for group in self.g_checked:
            self.button_dict[group] = self.chb[group][2]

    def show_hide(self, fr: Frame, name: str):
        """
        Toggles a frame to appear or disappear.
        :param fr: Frame to toggle
        :param name: name of the group. Used for a 'checked' label to appear or switch color
        """
        self.update_dict()
        on_off: BooleanVar = self.button_dict[name]
        self.g_checked[name].configure(text=" " * 42 + "[checked]")
        if on_off.get():
            self.g_checked[name].configure(fg='green')
            fr.pack(side="bottom", anchor="w")
            self.chb[name][1].configure(text=" Hide ")
        else:
            self.g_checked[name].configure(fg='grey')
            fr.pack_forget()
            self.chb[name][1].configure(text="Show")

    def browse_file(self, name: str, write: bool = False):
        """
        Opens a file dialog to browse for an existing file (read or write) or a new file (write).
        If write parameter is False, only existing files are allowed to choose.
        :param name: name of the corresponding parameter
        :param write: Enables non-existing files for writing
        """
        if write:
            file = filedialog.asksaveasfilename()
        else:
            file = filedialog.askopenfilename()
        if file:
            self.files_curr[name].configure(text=file, fg="black")
            self.files[name] = file

    def unset(self, name: str):
        """
        Sets the value of an FileType parameter to None. Mostly used for default setups of disabling functions.
        :param name: id of parameter (e.g. 'Ol').
        """
        self.files_curr[name].configure(text="Unset", fg="red")
        self.files[name] = None

    def adjust_coupling(self):
        """
        Opens coupling manager and refreshes coupling.
        """
        if len(self.coupling_info) == 0:
            ui = CouplingManager(descriptor=self.descriptor)
        else:
            ui = CouplingManager(descriptor=self.descriptor, coupled_res=self.coupling_info)
        self.coupling_info = ui.run_mode()

    def show_info(self, name: str):
        """
        Shows a messagebox with long id of 'name' and its help text
        :param name: name of parameter (e.g. atlas)
        """
        long_id: str = list(self.conversion_dict.keys())[list(self.conversion_dict.values()).index(name)]
        messagebox.showinfo('Info ' + long_id, help_txts[name])

    def insert_spinbox(self, name: str, frame_id: Frame, dec_dig: int, lim: Tuple = None):
        """
        Inserts a spinbox.
        :param self: ParameterManager object we're working in
        :param name: Name of the element (short version)
        :param frame_id: Frame to put widgets in
        :param dec_dig: Decimal digits that should be shown. Also alters the increment.
        :param lim: Tuple which includes minimum and maximum number for spinbox
        """
        self.insert_widget_basics(name, frame_id)
        default_num = IntVar()
        default_num.set(self.defaults[name])
        if lim is not None:
            start_ = lim[0]
            stop_ = lim[1]
        else:
            start_ = 0
            stop_ = 100
        self.spins[name] = Spinbox(frame_id, format="%." + str(dec_dig) + "f",
                                   increment=1.0 if dec_dig < 1 else 0.1,
                                   from_=start_, to=stop_, width=25,
                                   textvariable=default_num)
        self.spins[name].grid(column=2, row=self.row_)

    def insert_radio(self, name: str, frame_id: Frame):
        """
        Inserts several radio buttons. The number is based on the choices in self.radios_choices for 'name'.
        :param self: ParameterManager object we're working in
        :param name: Name of the element (short version)
        :param frame_id: Frame to put widgets in
        """
        self.insert_widget_basics(name, frame_id)
        self.radios_val[name] = StringVar()
        self.radios_val[name].set(self.defaults[name])
        self.radio_frames[name] = Frame(frame_id)
        for i in range(len(self.radios_choices[name])):
            self.radios[name][i] = Radiobutton(self.radio_frames[name],
                                               text=self.radios_choices[name][i],
                                               value=self.radios_choices[name][i],
                                               variable=self.radios_val[name])
            self.radios[name][i].grid(column=i, row=0)
        self.radio_frames[name].grid(column=2, row=self.row_, columnspan=10, sticky=W)

    def insert_checkbutton(self, name: str, frame_id: Frame):
        """
        Inserts a checkbutton.
        :param self: ParameterManager object we're working in
        :param name: Name of the element (short version)
        :param frame_id: Frame to put widgets in
        """
        self.insert_widget_basics(name, frame_id)
        self.checks_val[name] = BooleanVar()
        self.checks_val[name].set(self.defaults[name])
        self.checks[name] = Checkbutton(frame_id, text='Choose', var=self.checks_val[name])
        self.checks[name].grid(column=2, row=self.row_)

    def insert_coupling_option(self, frame_id: Frame):
        """
        Inserts button which triggers a coupling manager in a separate window.
        :param frame_id: Frame to put button in
        """
        coupling_button = Button(frame_id, text="Adjust coupling", command=self.adjust_coupling)
        coupling_button.grid(column=3, row=self.row_, sticky=W)

    def insert_combobox(self, name: str, frame_id: Frame, *args: str):
        """
        Inserts a Combobox.
        :param self: ParameterManager object we're working in
        :param name: Name of the element (short version)
        :param frame_id: Frame to put widgets in
        :param args: Additional choices for combobox value selection
        """
        self.insert_widget_basics(name, frame_id)
        default = self.defaults[name] if self.defaults[name] is not None else "value unset"
        values_ = [default]
        for x in args:
            values_.append(str(x))
        self.entrys[name] = ttk.Combobox(frame_id,
                                         text=name, values=values_, width=25)
        self.entrys[name].set(default)  # set the selected item
        self.entrys[name].grid(column=2, row=self.row_)

    def insert_filebox(self, name: str, frame_id: Frame, write_file: bool = False):
        """
        Inserts a Filebox with a browse button.
        :param self: ParameterManager object we're working in
        :param name: Name of the element (short version)
        :param frame_id: Frame to put widgets in
        :param write_file: Non-existing files are allowed, if enabled
        """
        self.insert_widget_basics(name, frame_id)
        # Change defaults to strings (from None or TextIOWrapper)
        # if self.defaults[name] is None:
        #    self.defaults[name] = "None"
        if isinstance(self.defaults[name], io.IOBase):
            self.defaults[name] = self.defaults[name].name
        if self.defaults[name] is None:
            self.files_curr[name] = self.insert_element_label(name, frame_id, True, column=2,
                                                              text="Unset", fg="red")
        else:
            self.files_curr[name] = self.insert_element_label(name, frame_id, True, column=2,
                                                              text=self.defaults[name])
        if write_file:
            button = Button(frame_id, text="Save as...",
                            command=lambda: self.browse_file(name, write_file))
        else:
            button = Button(frame_id, text="Browse",
                            command=lambda: self.browse_file(name))
        self.files[name] = self.defaults[name]  # set the default selected item
        button.grid(column=3, row=self.row_)
        if arg[name]['type'] in ("FileType('w')", "FileType('r')") and arg[name]['default'] in ("", "genmut.log"):
            unset = Button(frame_id, text="unset",
                           command=lambda: self.unset(name))
            unset.grid(column=4, row=self.row_)

    def insert_major_frame(self, name: str, description: str):
        """
        Inserts a major frame which contents a frame which can be enabled to show parameter selection.
        This double frame insertion is needed to get the arranger structure: grid(pack(grid)).
        :param self: ParameterManager object we're working in
        :param name: Name of the group (short version)
        :param description: Description of the group
        """
        self.maj_frames[name][0] = Frame(self.window)
        self.maj_frames[name][0].pack(side="top", fill="x")
        self.maj_frames[name][2] = Frame(self.maj_frames[name][0])
        self.maj_frames[name][2].pack(side="top", anchor="w", fill="both", expand="true")
        self.chb[name][0] = self.insert_element_label(name, self.maj_frames[name][2], ret=True, sticky="w",
                                                      text=description, side="left", pack=True)
        self.chb[name][2] = BooleanVar()
        self.chb[name][2].set(False)  # set check state
        self.maj_frames[name][1] = Frame(self.maj_frames[name][0])
        self.chb[name][1] = Checkbutton(self.maj_frames[name][2], text='Show', var=self.chb[name][2],
                                        command=lambda: self.show_hide(self.maj_frames[name][1], name),
                                        indicatoron="false")
        self.chb[name][1].pack(side="right", anchor="w")
        self.g_checked[name] = self.insert_element_label(name, self.maj_frames[name][2], ret=True, text=" " * 61,
                                                         sticky="e", pack=True, side="right")

    def insert_group(self, id_: str):
        """
        Adds all elements of a parameter group. Widget type is chosen by type key value in 'arg' dict.
        :param self: ParameterManager object we're working in
        :param id_: Name of the group (short version)
        """
        self.row_ = 0
        frame = self.maj_frames[id_][1]
        for ar in arg:
            if arg[ar]['group'] == id_:
                if arg[ar]['type'] in ('float', 'int'):
                    if 'limit' in arg[ar]:
                        self.insert_spinbox(ar, frame, 0 if arg[ar]['type'] == 'int' else 2, arg[ar]['limit'])
                    else:
                        self.insert_spinbox(ar, frame, 0 if arg[ar]['type'] == 'int' else 2)
                    self.row_ = next_row(self.row_)
                elif ar != "G" and arg[ar]['type'] == 'bool':
                    self.insert_checkbutton(ar, frame)
                    if ar == "Mu":
                        self.insert_coupling_option(frame)
                    self.row_ = next_row(self.row_)
                elif arg[ar]['type'] == 'str' and arg[ar]['choices'] != "":
                    self.insert_radio(ar, frame)
                    self.row_ = next_row(self.row_)
                elif arg[ar]['type'] == 'str':
                    if "alter" in arg[ar]:
                        self.insert_combobox(ar, frame, arg[ar]['alter'])
                    else:
                        self.insert_combobox(ar, frame)
                    self.row_ = next_row(self.row_)
                elif arg[ar]['type'] == "FileType('r')":
                    self.insert_filebox(ar, frame)
                    self.row_ = next_row(self.row_)
                elif arg[ar]['type'] == "FileType('w')":
                    self.insert_filebox(ar, frame, True)
                    self.row_ = next_row(self.row_)

    def build_main_window(self):
        """
        This method sets up all the different Elements/Widgets in group specific frames.
        The frames can be en- as well as disabled by a checkbox.
        All Elements contain a Help box to display the help text and a Label for the short name.
        """

        # Mandatory parameters (atlas, scaffold)
        self.insert_major_frame('mp', "Mandatory parameters: ")
        fr_mp = self.maj_frames['mp'][1]
        self.row_ = 0
        for a in ("atlas", "scaffold"):
            self.insert_filebox(a, fr_mp)
            self.row_ = next_row(self.row_)

        # Ligand residues (modifies descriptor!)
        self.insert_major_frame('aa', "Peptide residue selection: ")
        fr_aa = self.maj_frames['aa'][1]
        for res in range(self.lig_length):
            self.residues[res] = ttk.Combobox(fr_aa, width=4, state="readonly")
            self.residues[res]['values'] = aa
            self.residues[res].current(aa.index(self.descriptor.ligand_info[res][2]))
            self.residues[res].grid(column=res, row=0)

        # Life cycle parameters
        self.insert_major_frame('lc', "Life cycle parameters: ")
        self.insert_group('lc')

        # Fitness function parameters
        self.insert_major_frame('fit', "Fitness function parameters: ")
        self.insert_group('fit')

        # Mutation parameters
        self.insert_major_frame('mut', "Mutation parameters: ")
        self.insert_group('mut')

        # Pmatrix scoring parameters
        self.insert_major_frame('sc', "Pmatrix scoring parameters: ")
        self.insert_group('sc')

        # Clash scoring parameters
        self.insert_major_frame('cs', "Clash scoring parameters: ")
        self.insert_group('cs')

        # Deep atlas parameters
        self.insert_major_frame('da', "Deep atlas parameters: ")
        self.insert_group('da')

        # Rosetta parameters
        self.insert_major_frame('rr', "Rosetta parameters: ")
        self.insert_group('rr')

        # Osprey parameters
        self.insert_major_frame('os', "Osprey parameters: ")
        self.insert_group('os')

        # Output options
        self.insert_major_frame('out', "output options: ")
        self.insert_group('out')

        save_ = Button(self.window, text="Save parameters", command=self.save_params)
        save_.pack(side="left")
        import_ = Button(self.window, text="Import parameters", command=self.import_params)
        import_.pack(side="left")

        self.insert_sendcancel()

        if self.atpar:
            self.import_params(self.atpar)
            self.atpar = False

        self.window.mainloop()


class CouplingManager(Interface):
    """
    An interface type class that defines a coupling manager window.
    This windows allows the user to define coupling individually and send it to the genetic mutator (by the parameter
    manager).
    """

    def __init__(self, descriptor: ComplexDescriptor, coupled_res=None):
        super().__init__("Mutation coupling manager", main=False)
        self.coupler: Dict[int, Dict[int, Any]] = {}
        self.descriptor = descriptor
        if coupled_res is None:  # Coupling manager is opened for first time -> coupled_res is read from descriptor
            self.coupled_res: List[List[str]] = self.read_descriptor("coupled_res")
        else:  # Coupling manager was open before -> coupled_res is read from coupled_res parameter
            cs: List[List[str]] = []
            for c in coupled_res:
                cs.append(list(filter(lambda a: a != "", c[1])))
            self.coupled_res = normalize_couples(cs)
        self.binder_res: List[str] = self.read_descriptor("binder_res")
        self.max_couplers = max(len(l) for l in self.coupled_res)
        self.coupling_info = self.descriptor.coupling_info
        self.cp_frame = None
        self.cp_daughter: Dict[int, Frame] = {}

    def run_mode(self) -> List[Tuple[str, List[str]]]:
        """
        builds the main window (coupling manager window) and returns coupling info, if main window is destroyed.
        :return: couling information which was defined in read_current_coupling
        """
        self.build_main_window()
        return self.coupling_info

    def read_descriptor(self, mode: str) -> List[List[str]] or List[str]:
        """
        Reads descriptor and saves binder or coupling info depending on mode (either "coupled_res" or "binder_res").
        :param mode: decides to return binder or coupling info
        :return: descriptor info depending on mode
        """
        if mode == "coupled_res":
            couples: List[List[str]] = []
            for couple in self.descriptor.coupling_info:
                couples.append(couple[1])

            couples = normalize_couples(couples)
            return couples
        elif mode == "binder_res":
            binder_res: List[str] = []
            for binder in self.descriptor.binder_info:
                binder_res.append(binder[1])
            binder_res.insert(0, "")
            return binder_res

    def read_current_coupling(self) -> bool:
        """
        Reads current status of Comboboxes that define coupling and combines it to self.coupling_info.
        If one residue number is found twice False is returned, otherwise True.
        :return: False if one residue id is taken twice, else True.
        """

        def remove_empty_lines(cci: List[List[str or List[str]]]) -> List[List[str or List[str]]]:
            """
            Removes empty rows or columns (filled with "") when sending coupling information.
            :param cci: current coupling info which will be used by ParameterManager
            :return: cci list without empty lines
            """
            line_to_remove = []
            for line in range(len(cci)):
                if all(element == "" for element in cci[line][1]):
                    line_to_remove.append(line)
            for ltr in sorted(line_to_remove, reverse=True):
                del cci[ltr]
            col_to_remove = []
            for col in range(len(cci[0][1])):
                if all(cci[element][1][col] == "" for element in range(len(cci))):
                    col_to_remove.append(col)
            for ctr in sorted(col_to_remove, reverse=True):
                for line in range(len(cci)):
                    del cci[line][1][ctr]
            return cci

        doublecheck = set()
        current_coupling_info: List[List[str, List[str]]] = [None for _ in range(len(self.coupler))]
        for i in self.coupler:
            s = [None for _ in range(len(self.coupler[i]))]
            if len(self.descriptor.coupling_info) > 0:
                # Only chain of first couple is used
                row: List[str or List[str]] = [self.descriptor.coupling_info[0][0], s]
            else:  # If no couples are there use chain of first binder
                row: List[str or List[str]] = [self.descriptor.binder_info[0][0], s]
            for j in self.coupler[i]:
                row[1][j] = self.coupler[i][j].get()
                if row[1][j] not in doublecheck:
                    doublecheck.add(row[1][j])
                elif row[1][j] != "":
                    print("One residues is coupled twice!")
                    messagebox.showinfo('Sending not possible', "One residues is coupled twice!")
                    return False
            current_coupling_info[i] = row
        current_coupling_info = remove_empty_lines(current_coupling_info)
        for i in range(len(current_coupling_info)):
            current_coupling_info[i] = tuple(current_coupling_info[i])
        self.coupling_info = current_coupling_info
        return True

    def send_cmd(self):
        """
        Reads current coupling status and destroys the window.
        Is used when "send" button is hit to save coupling into the ParameterManager
        """
        print("Refreshing coupling data..")
        if self.read_current_coupling():
            self.window.destroy()

    def add_del_cr(self, name: str):
        """
        Adds or deletes a row or column in coupling matrix.
        :param name: defines the behaviour. There are four possibilities: "+col", "-col", "+row" or "-row".
        :return:
        """

        def ad_row(_self, add_or_del: str):
            """
            For adding or deleting a row.
            :param _self: self CouplingManager object instance
            :param add_or_del: "+" or "-"
            """
            if add_or_del == "+":
                m = max(_self.coupler.keys())
                d = [None for _ in _self.coupled_res[m]]
                for i, j in enumerate(_self.coupled_res[m]):
                    d[i] = ""
                _self.coupled_res.append(d)
                _self.load_couples(_self.cp_frame, m + 1)

            elif add_or_del == "-":
                m = max(_self.coupler.keys())
                if m >= 1:
                    _self.cp_daughter[m].destroy()
                    del _self.coupler[m]
                    del _self.cp_daughter[m]
                    del _self.coupled_res[m]
                # del _self.coupler[m]

        def ad_col(_self, add_or_del: str):
            """
            For adding or deleting a column.
            :param _self: self CouplingManager object instance
            :param add_or_del: "+" or "-"
            """
            if add_or_del == "+":
                bb = _self.binder_res
                m = max(_self.coupler[0].keys()) + 1  # works if at least one line is there
                for i in range(len(_self.coupled_res)):
                    _self.coupler[i][m] = ttk.Combobox(_self.cp_daughter[i], width=4, state="readonly")
                    _self.coupler[i][m]['values'] = bb
                    _self.coupler[i][m].current(0)
                    _self.coupler[i][m].grid(column=m + 1, row=0)
                    _self.coupled_res[i].append("")
                _self.max_couplers += 1

            elif add_or_del == "-":
                m = max(_self.coupler[0].keys())
                if m > 1:
                    for i in range(len(_self.coupled_res)):
                        _self.coupled_res[i].pop()
                        _self.coupler[i][m].destroy()
                        del _self.coupler[i][m]
                    _self.max_couplers -= 1

        pm = name[0]
        cr = name[1:]
        if cr == "row":
            ad_row(self, pm)
        elif cr == "col":
            ad_col(self, pm)
        else:
            sys.exit()

    def insert_plusminus(self, frame_id: Frame, row_or_column: str, plus_or_minus: str):
        """
        Inserts a button in desired frame. This button either adds or deletes a row or column in coupling matrix.
        The behaviour is defined in row_or_column or plus_or_minus.
        :param frame_id: The frame it should be placed in
        :param row_or_column: Defines the axis the button should work on ("col" columns or "row" rows in matrix)
        :param plus_or_minus: Defines the action. Either "+" for adding or "-" for deleting a line.
        """
        if row_or_column == "col":
            side = "right"
            text = " " + plus_or_minus + " "
        elif row_or_column == "row":
            side = "bottom"
            text = plus_or_minus
        else:
            sys.exit()
        info_button = Button(frame_id, text=text, fg='white', bg='bisque', font=self.bold,
                             command=lambda: self.add_del_cr(plus_or_minus + row_or_column))
        info_button.pack(side=side)

    def load_couples(self, frame: Frame or Toplevel, single: int = None):
        """
        Loads coupling matrix which harbours rows and columns that define residues groups which shall be coupled.
        By default all residues in self.coupled_res (defined from descriptor in __init__()) are loaded and
        build in the window.
        If single is stated, only this single row position is reloaded (e.g. if a new row is created with '+' button)
        :param frame: Frame the matrix should be placed in
        :param single: Defines single row to (re)build
        """
        bb = self.binder_res
        if single is None:
            rows = range(len(self.coupled_res))
        else:
            rows = range(single, single + 1)
        for i in rows:
            self.cp_daughter[i] = Frame(frame)
            self.cp_daughter[i].pack(side="top")
            num = str(i + 1) if len(str(i + 1)) >= 2 else "0" + str(i + 1)
            self.insert_element_label(f"Couple #{num}", self.cp_daughter[i], row=0, column=0)
            self.coupler[i] = {}
            for j in range(self.max_couplers):
                self.coupler[i][j] = ttk.Combobox(self.cp_daughter[i], width=4, state="readonly")
                self.coupler[i][j]['values'] = bb
                self.coupler[i][j].current(bb.index(self.coupled_res[i][j]))
                self.coupler[i][j].grid(column=j + 1, row=0)

    def build_main_window(self):
        """
        Builds the main window - in this case a coupling managing window.
        The window will last and block commands after this one until it's destroyed.
        """
        main_top = Frame(self.window)
        main_top.pack(side="top", fill="x", expand=False)
        main_bottom = Frame(self.window)
        main_bottom.pack(side="bottom", fill="both", expand=True)
        main_bottom_top = Frame(main_bottom)
        main_bottom_top.pack(side="top", fill="both", expand=True)

        main_left = Frame(main_bottom_top)
        main_left.pack(side="left", fill="y", expand=False)
        main_right = Frame(main_bottom_top)
        main_right.pack(side="right", fill="both", expand=True)

        cols_left = Frame(main_top)
        cols_left.pack(side="left", expand=False)
        cols_right = Frame(main_top)
        cols_right.pack(side="right", fill="x", expand=True)
        self.insert_help_button("help", cols_left)
        for i, pm in enumerate(["+", "-"]):
            self.insert_plusminus(cols_right, "col", pm)
        rows_ = Frame(main_left)
        rows_.pack(side="bottom", fill="y", expand=True)
        for i, pm in enumerate(["+", "-"]):
            self.insert_plusminus(rows_, "row", pm)

        self.cp_frame = Frame(main_right)
        self.load_couples(self.cp_frame)
        self.insert_sendcancel(frame=main_bottom)
        self.cp_frame.pack(side="left", fill="both", expand=False)
        self.window.wait_window()



from argparse import FileType, _ArgumentGroup, ArgumentParser
from typing import Dict

# Every dict needs a type. Type bool will only be taken for identification, not used in argparse.
# Every int, float and str type dict need a default value. str and FileType('r') defaults can be "", i.e. no default.
# Every str type dict needs choices. empty string ("") means there are no given choices.
# 'alter' key is used for alternative but not mandatory choices and is only considered if choices is "".
atlas_dict = {"group": "None", "type": "FileType('r')"}
scaffold_dict = {"group": "None", "type": "FileType('r')"}
descriptor_dict = {"group": "None", "type": "FileType('r')"}
g_dict = {"group": 'opt None', "l_name": 'graphical_input', "type": 'bool', "action": "store_true"}
a_dict = {"group": 'opt None', "l_name": 'atpar', "type": 'str', "default": "", "choices": ""}
lb_dict = {"group": 'lc', "l_name": 'max_siblings', "type": 'int', "default": 4}
lc_dict = {"group": 'lc', "l_name": 'n_clones', "type": 'int', "default": 1}
le_dict = {"group": 'lc', "l_name": 'n_epochs', "type": 'int', "default": 10}
lg_dict = {"group": 'lc', "l_name": 'max_generations', "type": 'int', "default": 10}
lg1_dict = {"group": 'lc', "l_name": 'gen_first_epoch', "type": 'int', "default": -1}
lk_dict = {"group": 'lc', "l_name": 'n_workers', "type": 'int', "default": -1, "limit": (-1, 128)}
ln_dict = {"group": 'lc', "l_name": 'n_top', "type": 'int', "default": 20}
lr_dict = {"group": 'lc', "l_name": 'n_recomb', "type": 'int', "default": 8}
fs_dict = {"group": 'fit', "l_name": 'scorer', "type": 'str', "default": 'rosetta',
           "choices": ["rosetta", "osprey", "simulate"]}
mc_dict = {"group": 'mut', "l_name": 'clustering_factor', "type": 'float', "default": 1.0}
md_dict = {"group": 'mut', "l_name": 'deep_factor', "type": 'float', "default": 0.0}
mf_dict = {"group": 'mut', "l_name": 'clash_factor', "type": 'float', "default": 1.0}
mr_dict = {"group": 'mut', "l_name": 'random_mutations', "type": 'bool', "action": "store_true"}
mrc_dict = {"group": 'mut', "l_name": 'random_mutation_chance', "type": 'int', "default": 20, "limit": (0, 100)}
mu_dict = {"group": 'mut', "l_name": 'coupling', "type": 'bool', "action": "store_true"}
sd_dict = {"group": 'sc', "l_name": 'distance_factor', "type": 'float', "default": 1.0}
se_dict = {"group": 'sc', "l_name": 'score_exponent', "type": 'float', "default": 2.0}
so_dict = {"group": 'sc', "l_name": 'orient_factor', "type": 'float', "default": 2.0}
ss_dict = {"group": 'sc', "l_name": 'secor_factor', "type": 'float', "default": 2.0}
st_dict = {"group": 'sc', "l_name": 'score_threshold', "type": 'float', "default": 8.0}
cc_dict = {"group": 'cs', "l_name": 'clash_merge_base', "type": 'float', "default": 1.5}
cd_dict = {"group": 'cs', "l_name": 'delta_factor', "type": 'float', "default": 4.0}
cn_dict = {"group": 'cs', "l_name": 'n_delta', "type": 'float', "default": 5.0}
cs_dict = {"group": 'cs', "l_name": 'std_dev_weight', "type": 'float', "default": 1.5}
cq_dict = {"group": 'cs', "l_name": 'n_quality_k', "type": 'float', "default": 5.0}
da_dict = {"group": 'da', "l_name": 'deep_alpha', "type": 'float', "default": 0.01}
dc_dict = {"group": 'da', "l_name": 'deep_activation', "type": 'str', "default": 'tanh',
           "choices": ['identity', 'logistic', 'tanh', 'relu']}
dd_dict = {"group": 'da', "l_name": 'deep_distance_cutoff', "type": 'float', "default": 8.0}
dh_dict = {"group": 'da', "l_name": 'deep_hidden_neurons', "type": 'int', "default": 25}
dl_dict = {"group": 'da', "l_name": 'deep_learning_rate', "type": 'float', "default": 0.005}
dm_dict = {"group": 'da', "l_name": 'deep_max_iter', "type": 'int', "default": 500}
dt_dict = {"group": 'da', "l_name": 'deep_tolerance', "type": 'float', "default": 1e-6}
dv_dict = {"group": 'da', "l_name": 'deep_verbose', "type": 'bool', "action": "store_true"}
rb_dict = {"group": 'rr', "l_name": 'rosetta_binding_site_factor', "type": 'float', "default": -1.0, "limit": (-1, 100)}
rd_dict = {"group": 'rr', "l_name": 'rosetta_db_path', "type": 'str', "default": "$ROSETTA_DB", "choices": ""}
re_dict = {"group": 'rr', "l_name": 'rosetta_extra_rotamers', "type": 'bool', "action": "store_true"}
rl_dict = {"group": 'rr', "l_name": 'rosetta_ligand_factor', "type": 'float', "default": 10.0}
rm_dict = {"group": 'rr', "l_name": 'rosetta_mode', "type": 'str', "default": 'relax',
           "choices": ['relax', 'score']}
rn_dict = {"group": 'rr', "l_name": 'rosetta_nstruct', "type": 'int', "default": 1}
ro_dict = {"group": 'rr', "l_name": 'rosetta_rmsd_offset', "type": 'float', "default": 3.0}
rp_dict = {"group": 'rr', "l_name": 'rosetta_path', "type": 'str', "default": "$ROSETTA", "choices": ""}
rq_dict = {"group": 'rr', "l_name": 'rosetta_pen_factor', "type": 'float', "default": 10.0}
rr_dict = {"group": 'rr', "l_name": 'rosetta_rmsd_slope', "type": 'float', "default": 2.0}
rs_dict = {"group": 'rr', "l_name": 'rosetta_relax_script', "type": "FileType('r')", "default": ""}
rv_dict = {"group": 'rr', "l_name": 'rosetta_negative_design', "type": 'bool', "action": "store_true"}
rx_dict = {"group": 'rr', "l_name": 'rosetta_relax_executable', "type": 'str',
           "default": 'relax.static.linuxgccrelease', "choices": ""}
ry_dict = {"group": 'rr', "l_name": 'rosetta_linearity_factor', "type": 'float', "default": 5.0}
pa_dict = {"group": 'os', "l_name": 'osprey_astar_mode', "type": 'str', "default": 'mplp',
           "choices": ['traditional', 'mplp']}
pe_dict = {"group": 'os', "l_name": 'osprey_use_epic', "type": 'bool', "action": "store_true"}
pf_dict = {"group": 'os', "l_name": 'osprey_finder', "type": 'str', "default": 'gmec', "choices": ['gmec', 'deegmec']}
pl_dict = {"group": 'os', "l_name": 'osprey_use_lute', "type": 'bool', "action": "store_true"}
ps_dict = {"group": 'os', "l_name": 'osprey_sidechain_flexibility', "type": 'str', "default": 'continuous',
           "choices": ['library', 'discrete', 'continuous']}
od_dict = {"group": 'out', "l_name": 'mutator_dumpfile', "type": 'str', "choices": "", "default": "",
           "alter": 'genmut.dump'}
ol_dict = {"group": 'out', "l_name": 'log_file', "type": "FileType('w')", "default": "genmut.log"}
om_dict = {"group": 'out', "l_name": 'mutant_dir', "type": 'str', "default": './mutants/', "choices": "", "alter": './'}
oo_dict = {"group": 'out', "l_name": 'output_file', "type": "FileType('w')", "default": 'genmut.score'}
op_dict = {"group": 'out', "l_name": 'mutant_pattern', "type": 'str', "default": 'mutant-$.pdb', "choices": ""}

arg: Dict[str, Dict] = {"atlas": atlas_dict,
                        "scaffold": scaffold_dict,
                        "descriptor": descriptor_dict,
                        "G": g_dict,
                        "A": a_dict,
                        "Lb": lb_dict,
                        "Lc": lc_dict,
                        "Le": le_dict,
                        "Lg": lg_dict,
                        "Lg1": lg1_dict,
                        "Lk": lk_dict,
                        "Ln": ln_dict,
                        "Lr": lr_dict,
                        "Fs": fs_dict,
                        "Mc": mc_dict,
                        "Md": md_dict,
                        "Mf": mf_dict,
                        "Mr": mr_dict,
                        "Mrc": mrc_dict,
                        "Mu": mu_dict,
                        "Sd": sd_dict,
                        "Se": se_dict,
                        "So": so_dict,
                        "Ss": ss_dict,
                        "St": st_dict,
                        "Cc": cc_dict,
                        "Cd": cd_dict,
                        "Cn": cn_dict,
                        "Cs": cs_dict,
                        "Cq": cq_dict,
                        "Da": da_dict,
                        "Dc": dc_dict,
                        "Dd": dd_dict,
                        "Dh": dh_dict,
                        "Dl": dl_dict,
                        "Dm": dm_dict,
                        "Dt": dt_dict,
                        "Dv": dv_dict,
                        "Rb": rb_dict,
                        "Rd": rd_dict,
                        "Re": re_dict,
                        "Rl": rl_dict,
                        "Rm": rm_dict,
                        "Rn": rn_dict,
                        "Ro": ro_dict,
                        "Rp": rp_dict,
                        "Rq": rq_dict,
                        "Rr": rr_dict,
                        "Rs": rs_dict,
                        "Rv": rv_dict,
                        "Rx": rx_dict,
                        "Ry": ry_dict,
                        "Pa": pa_dict,
                        "Pe": pe_dict,
                        "Pf": pf_dict,
                        "Pl": pl_dict,
                        "Ps": ps_dict,
                        "Od": od_dict,
                        "Ol": ol_dict,
                        "Om": om_dict,
                        "Oo": oo_dict,
                        "Op": op_dict}

help_txts: Dict[str, str] = \
    {"atlas": "the .atlas file to take as basis for genetic mutation.",
     "scaffold": "the PDB scaffold to optimize",
     "descriptor": "the descriptor file defining the ligand residue types as well as mutable host residues",
     "G": "Opens graphical interface for input arguments.",
     "A": "Load values from '.atpar' file. Only enabled in combination with graphical interface.",
     "Lb": "the maximum number of individuals with identical mutation sets that may coexist per epoch. Should be a "
           "multiple of n_clones",
     "Lc": "this number of clones is created for every individual generated. Clones are then processedin parallel, "
           "each is carrying a unique individual ID. Original is also counted.",
     "Le": "the number of epochs of genetic optimization",
     "Lg": "the number of recombining generations per epoch",
     "Lg1": "the number of recombining generations per epoch in first epoch.",
     "Lk": "Number of workers for parallel processing recombinations. -1 for number of CPU cores",
     "Ln": "the number of top elements to compare after each iteration. If the ranking obtained this way does not "
           "change, the script is aborted. Top mutations are also printed out regularly.",
     "Lr": "the number best mutants to recombine with base mutations in each generation",
     "Fs": "the scorer to use for calculating the fitness and the optimized structure of individuals",
     "Mc": "probability of distance clustering based mutation selection (default selection strategy)",
     "Md": "probability to involve a deep learning strategy for mutation selection",
     "Mf": "probability for clash factoring of the selection of mutations in each selection round",
     "Mr": "whether to apply random mutations.",
     "Mrc": "The chance of random mutations (over ones based on pmatrix). The chance is given in percent and only "
            "taken into account, if random mutations are enabled.",
     "Mu": "whether to apply analogous mutations to coupled residues automatically",
     "Sd": "weight factor for distance coefficient of the scoring function. [1/Angstrom]",
     "Se": "the outcome of the scoring function is postprocessed by raising by this value.",
     "So": "weight factor for the primary orientation coefficient (C_alpha->C_beta) of the scoring function. "
           "[1/radians]",
     "Ss": "weight factor for the secondary orientation coefficient (C_alpha->C_O) of the scoring function. "
           "[1/radians]",
     "St": "threshold coefficient of the scoring function for residue alignment during the pmatrix calculation "
           "in every epoch.",
     "Cc": "Base of clash punishment for merging with normalized quality: n_quality*BASE^-clash.",
     "Cd": "Correction factor for merging clash/quality with delta within calculation of mutation "
           "ranking: 'n_quality*base^-clash + n_delta/FACTOR'.",
     "Cn": "Normalization constant for delta. Equals to delta value of 50%% in normalized state.",
     "Cs": "Base in weighing the standard deviation: BASE^-std_dev * clash.",
     "Cq": "Normalization constant for quality. Equals to quality value of 50%% in normalized state.",
     "Da": "a coefficient that controls how sharp classification output will become (lower = sharper)",
     "Dc": "the activation function to use for the backpropagation algorithm while learning.",
     "Dd": "the minimum length of a chain to be treated as ligand",
     "Dh": "the size of the hidden layer in the underlying neural classificator.",
     "Dl": "learning rate of the neural classificator.",
     "Dm": "maximum number of iterations to process for each training data set",
     "Dt": "a coefficient that controls how small the training effect must be to terminate tranining",
     "Dv": "verbose training output",
     "Rb": "The share of binding site energy in the calculation of total energy by Rosetta relax. If -1, it is set"
           "to the ligand factor value.",
     "Rd": "Path or environment variable pointing to the Rosetta database folder",
     "Re": "Whether Rosetta relax uses extra rotamers (Rosetta flags -ex1 -ex2 set)",
     "Rl": "The share of ligand energy in the calculation of total energy by Rosetta relax",
     "Rm": "Rosetta's execution mode. Use 'simulate' for testing purposes only. 'score' will not relax the "
           "structure, but just score it, which is much faster yet unrealistic.",
     "Rn": "number of starting points for every Rosetta relax run. Hint: consider parameter n_clones instead, "
           "which behaves almost equivalently yet allows for parallel Rosetta runs.",
     "Ro": "An offset value where rmsd penalty will start",
     "Rp": "Path or environment variable pointing to the folder containing Rosetta executables",
     "Rq": "weight factor of the RMSD penalty",
     "Rr": "The slope of the influence of ligand rmsd after the offset",
     "Rs": "Rosetta relax script to use. If unset, use default relax:fast protocol. Ignored if rosetta_mode "
           "is 'simulate'.",
     "Rv": "Use negative design for best individuals to test specificity",
     "Rx": "Name of the relax executable in rosetta_path",
     "Ry": "Impact of peptide linearity on the calculation of total energy by Rosetta relax",
     "Pa": "A* implementation",
     "Pe": "whether to use the EPIC algorithm for continuous flexibility modeling",
     "Pf": "Finder method",
     "Pl": "whether to use the LUTE algorithm for continuous flexibility modeling",
     "Ps": "Resolution for sidechain optimization",
     "Od": "if this file exists, the genetic mutator will be reconstructed from it. Furthermore, after every "
           "generation, the current progress is dumped into this file.",
     "Ol": "if set, the log output is forwarded to this file, otherwise to stdout",
     "Om": "directory where mutant pdb and log files are stored. Will be created if not existant.",
     "Oo": "output file. Linewise output in format <total-score> <epoch> <generation> <mutant-id> <mutations>, "
           "ordered by ascending total-score.",
     "Op": "file name pattern for mutated structures. '$' will be replaced by the line number of the corresponding "
           "mutation in the mutations file."}


def add_argument(group: ArgumentParser or _ArgumentGroup, name: str, optional: bool = True):
    if optional:
        arg_name = "-" + name
        l_name = "--" + arg[name]['l_name']
        if arg[name]['type'] == 'bool':
            group.add_argument(arg_name, l_name, action=arg[name]['action'], help=help_txts[name])
        elif arg[name]['type'] == 'int':
            group.add_argument(arg_name, l_name, default=arg[name]['default'], type=int,
                               help=help_txts[name])
        elif arg[name]['type'] == 'float':
            group.add_argument(arg_name, l_name, default=arg[name]['default'], type=float,
                               help=help_txts[name])
        elif arg[name]['type'] == 'str':
            if arg[name]['choices'] != "":
                if arg[name]['default'] != "":
                    group.add_argument(arg_name, l_name, default=arg[name]['default'], type=str,
                                       choices=arg[name]['choices'], help=help_txts[name])
                else:
                    group.add_argument(arg_name, l_name, type=str,
                                       choices=arg[name]['choices'], help=help_txts[name])
            else:
                if arg[name]['default'] != "":
                    group.add_argument(arg_name, l_name, default=arg[name]['default'], type=str,
                                       help=help_txts[name])
                else:
                    group.add_argument(arg_name, l_name, type=str, help=help_txts[name])
        elif arg[name]['type'] == "FileType('r')":
            if arg[name]['default'] != "":
                group.add_argument(arg_name, l_name, default=arg[name]['default'], type=FileType('r'),
                                   help=help_txts[name])
            else:
                group.add_argument(arg_name, l_name, type=FileType('r'), help=help_txts[name])
        elif arg[name]['type'] == "FileType('w')":
            if arg[name]['default'] != "":
                group.add_argument(arg_name, l_name, default=arg[name]['default'], type=FileType('w'),
                                   help=help_txts[name])
            else:
                group.add_argument(arg_name, l_name, type=FileType('w'), help=help_txts[name])

    else:
        if arg[name]['type'] == 'bool':
            group.add_argument(name, action=arg[name]['action'], help=help_txts[name])
        elif arg[name]['type'] == 'int':
            group.add_argument(name, type=int, help=help_txts[name])
        elif arg[name]['type'] == 'float':
            group.add_argument(name, type=float, help=help_txts[name])
        elif arg[name]['type'] == 'str':
            if arg[name]['choices'] != "":
                group.add_argument(name, type=str, choices=arg[name]['choices'], help=help_txts[name])
            else:
                group.add_argument(name, type=str, help=help_txts[name])
        elif arg[name]['type'] == "FileType('r')":
            group.add_argument(name, type=FileType('r'), help=help_txts[name])
        elif arg[name]['type'] == "FileType('w')":
            group.add_argument(name, type=FileType('w'), help=help_txts[name])
    return group
